import logging

from typeguard import typechecked

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

from .logging_utils import _format_sql

from .gammas import (
    _add_left_right,
    _retain_source_dataset_column,
    _add_unique_id_and_source_dataset,
)
from .model import Model
from .ordered_set import OrderedSet
from .settings import ComparisonColumn

logger = logging.getLogger(__name__)


@typechecked
def run_expectation_step(df_with_gamma: DataFrame, model: Model, spark: SparkSession):
    """Run the expectation step of the EM algorithm described in the fastlink paper:
    http://imai.fas.harvard.edu/research/files/linkage.pdf

      Args:
          df_with_gamma (DataFrame): Spark dataframe with comparison vectors already populated
          model (Model): splink Model object
          spark (SparkSession): SparkSession

      Returns:
          DataFrame: Spark dataframe with a match_probability column
    """

    retain_source_dataset = _retain_source_dataset_column(
        model.current_settings_obj.settings_dict, df_with_gamma
    )

    sql = _sql_gen_gamma_bf_columns(model, retain_source_dataset)

    df_with_gamma.createOrReplaceTempView("df_with_gamma")
    logger.debug(_format_sql(sql))
    df_with_gamma_probs = spark.sql(sql)

    sql = _sql_gen_expected_match_prob(model, retain_source_dataset)

    logger.debug(_format_sql(sql))
    df_with_gamma_probs.createOrReplaceTempView("df_with_gamma_probs")
    df_e = spark.sql(sql)

    df_e.createOrReplaceTempView("df_e")

    model.save_settings_to_iteration_history()

    return df_e


def _sql_gen_gamma_bf_columns(
    model: Model, retain_source_dataset_col: bool, table_name="df_with_gamma"
):
    """
    For each row, look up the bayes factor associated with the observed
    gamma value
    """
    settings = model.current_settings_obj.settings_dict

    # Dictionary of case statements - these will be used in the list of columsn
    # in the SQL 'select' statement
    case_statements = {}
    for cc in model.current_settings_obj.comparison_columns_list:
        case_statements[f"bf_gamma_{cc.name}"] = _sql_gen_bayes_factors(cc)
        if cc.term_frequency_adjustments:
            case_statements[f"bf_tf_adj_{cc.name}"] = _sql_gen_bayes_factors(
                cc, tf_adj=True
            )

    select_cols = OrderedSet()
    uid = settings["unique_id_column_name"]
    sds = settings["source_dataset_column_name"]
    select_cols = _add_unique_id_and_source_dataset(
        select_cols, uid, sds, retain_source_dataset_col
    )

    for col in settings["comparison_columns"]:
        cc = ComparisonColumn(col)
        if settings["retain_matching_columns"]:
            for col_name in cc.columns_used:
                select_cols = _add_left_right(select_cols, col_name)

        if col["term_frequency_adjustments"]:
            select_cols.add(case_statements[f"bf_tf_adj_{cc.name}"])

        select_cols.add("gamma_" + cc.name)

        select_cols.add(case_statements[f"bf_gamma_{cc.name}"])

    for c in settings["additional_columns_to_retain"]:
        select_cols = _add_left_right(select_cols, c)

    if "blocking_rules" in settings:
        if len(settings["blocking_rules"]) > 1:
            select_cols.add("match_key")

    select_expr = ", ".join(select_cols)

    sql = f"""
    -- We use case statements for these lookups rather than joins for performance and simplicity
    select {select_expr}
    from {table_name}
    """

    return sql


def _column_order_df_e_select_expr(
    settings, retain_source_dataset_col, tf_adj_cols=False
):
    # Column order for case statement.  We want orig_col_l, orig_col_r, gamma_orig_col,
    select_cols = OrderedSet()
    uid = settings["unique_id_column_name"]
    sds = settings["source_dataset_column_name"]
    select_cols = _add_unique_id_and_source_dataset(
        select_cols, uid, sds, retain_source_dataset_col
    )

    for col in settings["comparison_columns"]:

        cc = ComparisonColumn(col)
        if settings["retain_matching_columns"]:
            for col_name in cc.columns_used:
                select_cols = _add_left_right(select_cols, col_name)

        select_cols.add(f"gamma_{cc.name}")

        if settings["retain_intermediate_calculation_columns"]:

            select_cols.add(f"bf_gamma_{cc.name}")

            if tf_adj_cols:
                if col["term_frequency_adjustments"]:
                    select_cols.add(f"bf_tf_adj_{cc.name}")

    for c in settings["additional_columns_to_retain"]:
        select_cols = _add_left_right(select_cols, c)

    if "blocking_rules" in settings:
        if len(settings["blocking_rules"]) > 1:
            select_cols.add("match_key")

    return ", ".join(select_cols)


def _sql_gen_expected_match_prob(
    model, retain_source_dataset, table_name="df_with_gamma_probs"
):
    settings = model.current_settings_obj.settings_dict
    ccs = model.current_settings_obj.comparison_columns_list

    λ = model.current_settings_obj["proportion_of_matches"]

    bf_cols = [f"bf_{cc.gamma_name}" for cc in ccs]
    bf_tf_adj_cols = [
        f"bf_tf_adj_{cc.name}" for cc in ccs if cc.term_frequency_adjustments
    ]
    bf_cols.extend(bf_tf_adj_cols)

    bayes_factor = " * ".join(bf_cols)
    bayes_factor = f"{λ}D/ (1-{λ}D) * {bayes_factor}"

    select_expr = _column_order_df_e_select_expr(
        settings, retain_source_dataset, tf_adj_cols=True
    )

    sql = f"""
    with df_with_overall_bf as (
        select
            *,
            {bayes_factor} as __bf
        from {table_name} )

    select
        log2(__bf) as match_weight,
        (__bf/(1+__bf)) as match_probability,
        {select_expr}
    from df_with_overall_bf
    """

    return sql


def _sql_gen_bayes_factors(comparison_column, tf_adj=False):
    """
    Create the case statements that look up the correct probabilities in the
    model dict for each gamma to calculate Bayes factors (m / u) and additional
    term frequency Bayes factors (u / (term frequency))
    """
    cc = comparison_column

    case_statements = []
    case_statements.append(f"when {cc.gamma_name} = -1 then 1.0D")

    if not tf_adj:
        alias = f"bf_{cc.gamma_name}"
        # This will fail if any of the values are null
        # If the values are null, this means that the value never occurs in the dataset
        # We can set the bf to null in this case

        m_u_zip = zip(cc["m_probabilities"], cc["u_probabilities"])

        for gamma_index, (m, u) in enumerate(m_u_zip):
            if m is None or u is None:
                case_stmt = f"when {cc.gamma_name} = {gamma_index} then null"
            else:
                case_stmt = f"when {cc.gamma_name} = {gamma_index} then {m/u}D"

            case_statements.append(case_stmt)
    else:
        alias = f"bf_tf_adj_{cc.name}"

        # The full weight tf adjustment is based on an exact match
        # i.e. it's computed based on how likely tokens are
        # to collide. relative to average
        # The adjustment is based on weight this adjustment down

        u_prob_exact = cc["u_probabilities"][-1]
        tf_weights = cc["tf_adjustment_weights"]

        case_stmt = f"when tf_{cc.name}_l is null or tf_{cc.name}_r is null then null"
        case_statements.append(case_stmt)

        for gamma_index, weight in enumerate(tf_weights):
            if u_prob_exact is not None:
                bf_tf = f"{u_prob_exact}D / greatest(tf_{cc.name}_l, tf_{cc.name}_r)"
                case_stmt = f"when {cc.gamma_name} = {gamma_index} then power({bf_tf}, {weight}D)"
            else:
                case_stmt = f"when {cc.gamma_name} = {gamma_index} then 1"

            case_statements.append(case_stmt)

    case_statements = "\n".join(case_statements)

    sql = f""" case
        {case_statements}
        end
        as {alias}
        """

    return sql.strip()



import pkgutil
import json
import os

altair_installed = True
try:
    import altair as alt
except ImportError:
    altair_installed = False


def load_chart_definition(filename):
    path = f"files/chart_defs/{filename}"
    data = pkgutil.get_data(__name__, path)
    schema = json.loads(data)
    return schema


def _load_multi_chart_template():
    path = "files/templates/multi_chart_template.txt"
    template = pkgutil.get_data(__name__, path).decode("utf-8")
    return template


def _load_external_libs():

    to_load = {
        "vega-embed": "files/external_js/vega-embed@6.14.2",
        "vega-lite": "files/external_js/vega-lite@4.17.0",
        "vega": "files/external_js/vega@5.17.1",
    }

    loaded = {}
    for k, v in to_load.items():
        script = pkgutil.get_data(__name__, v).decode("utf-8")
        loaded[k] = script

    return loaded


def altair_if_installed_else_json(chart_dict):
    if altair_installed:
        return alt.Chart.from_dict(chart_dict)
    else:
        return chart_dict


def _make_json(chart_or_dict):
    if altair_installed:
        return chart_or_dict.to_json(indent=None)
    else:
        return json.dumps(chart_or_dict)


iframe_message = """
To view in Jupyter you can use the following command:

from IPython.display import IFrame
IFrame(src="./{filename}", width=1000, height=500)
"""


def save_offline_chart(
    altair_chart, filename="my_chart.html", overwrite=False, print_msg=True
):

    if os.path.isfile(filename) and not overwrite:
        raise ValueError(
            f"The path {filename} already exists. Please provide a different path."
        )

    # get altair chart as json
    path = "files/templates/single_chart_template.txt"
    template = pkgutil.get_data(__name__, path).decode("utf-8")

    fmt_dict = _load_external_libs()

    fmt_dict["mychart"] = _make_json(altair_chart)

    with open(filename, "w") as f:
        f.write(template.format(**fmt_dict))

    if print_msg:
        print(f"Chart saved to {filename}")
        print(iframe_message.format(filename=filename))



import logging
from .ordered_set import OrderedSet
from .settings import ComparisonColumn

from typeguard import typechecked

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

from .logging_utils import _format_sql

logger = logging.getLogger(__name__)


def sql_gen_comparison_columns(columns: list) -> str:
    """Build SQL expression that renames columns and sets them aside each other for comparisons

     Args:
         columns (list): List of cols

    Examples:
         >>> sql_gen_comparison_columns(["name", "dob"])
         "name_l, name_r, dob_l, dob_r"

     Returns:
         SQL expression
    """

    l = [f"l.{c} as {c}_l" for c in columns]
    r = [f"r.{c} as {c}_r" for c in columns]
    both = zip(l, r)
    flat_list = [item for sublist in both for item in sublist]
    return ", ".join(flat_list)


def _retain_source_dataset_column(settings_dict, df):
    # Want to retain source dataset column in all cases
    # except when link type is dedupe only
    # and the column does not exist in the data
    if settings_dict["link_type"] != "dedupe_only":
        return True

    source_dataset_colname = settings_dict.get(
        "source_dataset_column_name", "source_dataset"
    )
    if source_dataset_colname in df.columns:
        return True
    else:
        return False


def _get_columns_to_retain_blocking(settings_dict, df):

    columns_to_retain = OrderedSet()
    columns_to_retain.add(settings_dict["unique_id_column_name"])
    if _retain_source_dataset_column(settings_dict, df):
        columns_to_retain.add(settings_dict["source_dataset_column_name"])
    for col in settings_dict["comparison_columns"]:
        cc = ComparisonColumn(col)
        for col_name in cc.columns_used:
            columns_to_retain.add(col_name)
        if cc["term_frequency_adjustments"] and f"tf_{cc.name}" in df.columns:
            columns_to_retain.add(f"tf_{cc.name}")

    for c in settings_dict["additional_columns_to_retain"]:
        columns_to_retain.add(c)

    return columns_to_retain.as_list()


def _sql_gen_and_not_previous_rules(previous_rules: list):
    if previous_rules:
        # Note the isnull function is important here - otherwise
        # you filter out any records with nulls in the previous rules
        # meaning these comparisons get lost
        or_clauses = [f"ifnull(({r}), false)" for r in previous_rules]
        previous_rules = " OR ".join(or_clauses)
        return f"AND NOT ({previous_rules})"
    else:
        return ""


def _sql_gen_composite_unique_id(source_dataset_colname, unique_id_colname, l_or_r):
    return f"concat({l_or_r}.{source_dataset_colname}, '-__-',{l_or_r}.{unique_id_colname})"


def _sql_gen_where_condition(link_type, source_dataset_col, unique_id_col):

    if source_dataset_col:
        id_expr_l = _sql_gen_composite_unique_id(source_dataset_col, unique_id_col, "l")
        id_expr_r = _sql_gen_composite_unique_id(source_dataset_col, unique_id_col, "r")
    else:
        id_expr_l = f"l.{unique_id_col}"
        id_expr_r = f"r.{unique_id_col}"

    if link_type in ["link_and_dedupe", "dedupe_only"]:
        where_condition = f"where {id_expr_l} < {id_expr_r}"
    elif link_type == "link_only":
        where_condition = f"where {id_expr_l} < {id_expr_r} and l.{source_dataset_col} != r.{source_dataset_col}"

    return where_condition


def _sql_gen_block_using_rules(
    link_type: str,
    columns_to_retain: list,
    blocking_rules: list,
    unique_id_col: str = "unique_id",
    source_dataset_col: str = "source_dataset",
    table_name: str = "df",
):
    """Build a SQL statement that implements a list of blocking rules.

    The left and right tables are aliased as `l` and `r` respectively, so an example
    blocking rule would be `l.surname = r.surname AND l.forename = r.forename`.

    Args:
        link_type: One of 'link_only', 'dedupe_only', or 'link_and_dedupe'
        columns_to_retain: List of columns to keep in returned dataset
        blocking_rules: Each element of the list represents a blocking rule
        unique_id_col (str, optional): The name of the column containing the row's unique_id. Defaults to "unique_id".
        table_name (str, optional): Name of the table to dedupe (where (where `link_type` is `dedupe_only`). Defaults to "df".

    Returns:
        str: A SQL statement that implements the blocking rules
    """

    sql_select_expr = sql_gen_comparison_columns(columns_to_retain)
    where_condition = _sql_gen_where_condition(
        link_type, source_dataset_col, unique_id_col
    )

    sqls = []
    previous_rules = []
    for matchkey_number, rule in enumerate(blocking_rules):
        not_previous_rules_statement = _sql_gen_and_not_previous_rules(previous_rules)
        sql = f"""
        select
        {sql_select_expr},
        '{matchkey_number}' as match_key
        from {table_name} as l
        inner join {table_name} as r
        on
        {rule}
        {not_previous_rules_statement}
        {where_condition}
        """
        previous_rules.append(rule)
        sqls.append(sql)

    sql = "union all".join(sqls)

    return sql


def _sql_gen_cartesian_block(
    link_type: str,
    columns_to_retain: list,
    unique_id_col: str = "unique_id",
    source_dataset_col: str = "source_dataset",
    table_name: str = "df",
):
    """Build a SQL statement that performs a cartesian join.

    Args:
        link_type: One of 'link_only', 'dedupe_only', or 'link_and_dedupe'
        columns_to_retain: List of columns to keep in returned dataset
        unique_id_col (str, optional): The name of the column containing the row's unique_id. Defaults to "unique_id".

    Returns:
        str: A SQL statement that implements the join
    """

    sql_select_expr = sql_gen_comparison_columns(columns_to_retain)

    where_condition = _sql_gen_where_condition(
        link_type, source_dataset_col, unique_id_col
    )

    sql = f"""
    select
    {sql_select_expr},
    '0' as match_key
    from {table_name} as l
    cross join {table_name} as r
    {where_condition}
    """

    return sql


@typechecked
def block_using_rules(settings: dict, df: DataFrame, spark: SparkSession):
    """Apply a series of blocking rules to create a dataframe of record comparisons. If no blocking rules provided, performs a cartesian join.

    Args:
        settings (dict): A splink settings dictionary
        df (DataFrame): Spark dataframe to block - if linking multiple datasets, assumes dataframes have already been vertically concatenated
        spark (SparkSession): The pyspark.sql.session.SparkSession

    Returns:
        pyspark.sql.dataframe.DataFrame: A dataframe of each record comparison
    """
    df.createOrReplaceTempView("df")
    columns_to_retain = _get_columns_to_retain_blocking(settings, df)
    unique_id_col = settings["unique_id_column_name"]
    if settings["link_type"] == "dedupe_only":
        source_dataset_col = None
    else:
        source_dataset_col = settings["source_dataset_column_name"]
    link_type = settings["link_type"]

    if "blocking_rules" not in settings or len(settings["blocking_rules"]) == 0:
        sql = _sql_gen_cartesian_block(
            link_type, columns_to_retain, unique_id_col, source_dataset_col
        )
    else:
        rules = settings["blocking_rules"]
        sql = _sql_gen_block_using_rules(
            link_type, columns_to_retain, rules, unique_id_col, source_dataset_col
        )

    logger.debug(_format_sql(sql))

    df_comparison = spark.sql(sql)

    return df_comparison



import warnings
import re
from .jar_location import similarity_jar_location


def _check_jaro_registered(spark):

    if spark is None:
        return False

    if (
        spark == "supress_warnings"
    ):  # Allows us to surpress the warning in the test suite
        return False

    for fn in spark.catalog.listFunctions():
        if fn.name == "jaro_winkler_sim":
            return True

    warnings.warn(
        "\n\nCustom string comparison functions such as jaro_winkler_sim are not available in Spark\n"
        "Or you did not pass 'spark' (the SparkSession) into 'Model' \n"
        "You can import these functions using the scala-udf-similarity-0.0.9.jar provided with"
        " Splink.\n" + _get_spark_jars_string()
    )
    return False


def _get_spark_jars_string():
    """
    Outputs the exact string needed in the sparkSession config variable `spark.jars`
    In order to use the custom functions in the spark-udf-similarity-0.0.9.jar

    """

    path = similarity_jar_location()

    message = (
        "You will need to add it by correctly configuring your spark config\n"
        "For example in Spark 2.4.5\n"
        "\n"
        "from pyspark.sql import SparkSession, types\n"
        "from pyspark.context import SparkConf, SparkContext\n"
        f"conf.set('spark.driver.extraClassPath', '{path}') # Not needed in spark 3\n"
        f"conf.set('spark.jars', '{path}')\n"
        "spark.udf.registerJavaFunction('jaro_winkler_sim','uk.gov.moj.dash.linkage.JaroWinklerSimilarity',types.DoubleType())\n"
        "sc = SparkContext.getOrCreate(conf=conf)\n"
        "spark = SparkSession(sc)\n"
        "\n"
        "Alternatively, for Jaro Winkler, you can register a less efficient"
        " Python implementation using\n"
        "\n"
        "from splink.jar_fallback import jw_sim_py\n"
        "spark.udf.register('jaro_winkler_sim', jw_sim_py)\n"
    )

    return message


def _find_last_end_position(case_statement):
    # Since we're only interested in the position, case shouldn't matter.    stmt = case_statement.lower()
    case_statement = case_statement.lower()
    # we want to look for the final end that's delimited by whitespace (e.g. not as lender_name or something like that)
    regex = re.compile(r"\send\s?")
    try:
        m = list(re.finditer(regex, case_statement))[-1]
    except IndexError:
        raise ValueError(
            "Your case statement {case_statement} appears to be malformatted -"
            f" there's no END. Statement is {case_statement}"
        )
    start, end = m.span()

    return end


def _add_as_gamma_to_case_statement(case_statement: str, gamma_col_name):
    """As the correct column alias to the case statement if it does not exist

    If the case statement ends with no alias (case when....end) then adds it:
        case when...end as gamma_col_name

    If the case statement ends with an alias (case when...end as mycol) it replaces it:
        case when...end as gamma_col_name

    Args:
        case_statement (str): Original case statement

    Returns:
        str: case_statement with correct alias
    """
    if gamma_col_name is None:
        return case_statement

    # What we're trying to do is distinguish between 'case when end as gamma_blah' from case when end
    case_statement = case_statement.strip()

    # Strip out current alias if exists
    if case_statement[-4:].lower() != " end":
        # The case statement has a 'end as gamma blah'
        last_end_index = _find_last_end_position(case_statement)
        case_statement = case_statement[:last_end_index]
    return f"{case_statement} as gamma_{gamma_col_name}"


def _check_no_obvious_problem_with_case_statement(case_statement):
    seems_valid = True
    cs_l = case_statement.lower()
    if "case" not in cs_l:
        seems_valid = False
    if "end" not in cs_l:
        seems_valid = False
    if "when" not in cs_l:
        seems_valid = False
    if "then" not in cs_l:
        seems_valid = False

    if not seems_valid:
        s1 = "The case expression you provided does not seem to be valid SQL."
        s2 = f"Expression provided is: '{case_statement}'"
        raise ValueError(f"{s1} {s2}")


def sql_gen_case_smnt_strict_equality_2(col_name, gamma_col_name=None):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {col_name}_l = {col_name}_r then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


# String comparison select statements

# jaro winkler statements, using param values suggested on page 355 of
#  https://imai.fas.harvard.edu/research/files/linkage.pdf
# American Political Science Review (2019) 113, 2, 353–371, Using a Probabilistic Model to Assist Merging of Large-Scale
# Administrative Records


def sql_gen_case_stmt_jaro_2(col_name, threshold, gamma_col_name=None):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_jaro_3(
    col_name, gamma_col_name=None, threshold1=1.0, threshold2=0.88
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold1} then 2
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold2} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_jaro_4(
    col_name, gamma_col_name=None, threshold1=1.0, threshold2=0.88, threshold3=0.7
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold1} then 3
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold2} then 2
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold3} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_levenshtein_rel_3(
    col_name, gamma_col_name=None, threshold1=0.0, threshold2=0.3
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when levenshtein({col_name}_l, {col_name}_r)/((length({col_name}_l) + length({col_name}_r))/2) <= {threshold1} then 2
    when levenshtein({col_name}_l, {col_name}_r)/((length({col_name}_l) + length({col_name}_r))/2) <= {threshold2} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_levenshtein_abs_3(
    col_name, gamma_col_name=None, threshold1=0, threshold2=2
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {col_name}_l = {col_name}_r then 2
    when levenshtein({col_name}_l, {col_name}_r) <= {threshold1} then 2
    when levenshtein({col_name}_l, {col_name}_r) <= {threshold2} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_levenshtein_rel_4(
    col_name, gamma_col_name=None, threshold1=0.0, threshold2=0.2, threshold3=0.4
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {col_name}_l = {col_name}_r then 3
    when levenshtein({col_name}_l, {col_name}_r)/((length({col_name}_l) + length({col_name}_r))/2) <= {threshold1} then 3
    when levenshtein({col_name}_l, {col_name}_r)/((length({col_name}_l) + length({col_name}_r))/2) <= {threshold2}
    then 2
    when levenshtein({col_name}_l, {col_name}_r)/((length({col_name}_l) + length({col_name}_r))/2) <= {threshold3}
    then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_levenshtein_abs_4(
    col_name, gamma_col_name=None, threshold1=0, threshold2=1, threshold3=2
):
    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {col_name}_l = {col_name}_r then 3
    when levenshtein({col_name}_l, {col_name}_r) <= {threshold1} then 3
    when levenshtein({col_name}_l, {col_name}_r) <= {threshold2} then 2
    when levenshtein({col_name}_l, {col_name}_r) <= {threshold3} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


# Numeric value statements


def _sql_gen_max_of_two_cols(col1, col2):
    return f"""
    case
    when {col1} > {col2} then {col1}
    else {col2}
    end
    """


def _sql_gen_abs_diff(col1, col2):
    return f"(abs({col1} - {col2}))"


def sql_gen_case_stmt_numeric_float_equality_2(col_name, gamma_col_name=None):

    col1 = f"{col_name}_l"
    col2 = f"{col_name}_r"

    abs_difference = _sql_gen_abs_diff(col1, col2)

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {abs_difference} < 0.00001 then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_numeric_abs_3(
    col_name, gamma_col_name=None, abs_amount=1, equality_threshold=0.0001
):

    col1 = f"{col_name}_l"
    col2 = f"{col_name}_r"

    abs_difference = _sql_gen_abs_diff(col1, col2)

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {abs_difference} < {equality_threshold} THEN 2
    when {abs_difference} < {abs_amount} THEN 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_numeric_abs_4(
    col_name,
    gamma_col_name=None,
    abs_amount_low=1,
    abs_amount_high=10,
    equality_threshold=0.0001,
):

    col1 = f"{col_name}_l"
    col2 = f"{col_name}_r"

    abs_difference = _sql_gen_abs_diff(col1, col2)

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {abs_difference} < {equality_threshold} THEN 3
    when {abs_difference} < {abs_amount_low} THEN 2
    when {abs_difference} < {abs_amount_high} THEN 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_numeric_perc_3(
    col_name, gamma_col_name=None, per_diff=0.05, equality_threshold=0.0001
):

    col1 = f"{col_name}_l"
    col2 = f"{col_name}_r"

    max_of_cols = _sql_gen_max_of_two_cols(col1, col2)
    abs_difference = _sql_gen_abs_diff(col1, col2)

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {abs_difference}/abs({max_of_cols}) < {equality_threshold} then 2
    when {abs_difference}/abs({max_of_cols}) < {per_diff} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_numeric_perc_4(
    col_name,
    gamma_col_name=None,
    per_diff_low=0.05,
    per_diff_high=0.10,
    equality_threshold=0.0001,
):

    col1 = f"{col_name}_l"
    col2 = f"{col_name}_r"

    max_of_cols = _sql_gen_max_of_two_cols(col1, col2)
    abs_difference = _sql_gen_abs_diff(col1, col2)

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when {abs_difference}/abs({max_of_cols}) < {equality_threshold} then 3
    when {abs_difference}/abs({max_of_cols}) < {per_diff_low} then 2
    when {abs_difference}/abs({max_of_cols}) < {per_diff_high} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def _sql_gen_get_or_list_jaro(col_name, other_name_cols, threshold=1.0):
    # Note the ifnull 1234 just ensures that if one of the other columns is null,
    # the jaro score is lower than the threshold
    ors = [
        f"jaro_winkler_sim(ifnull({col_name}_l, '1234abcd5678'), ifnull({n}_r,"
        f" '987pqrxyz654')) >= {threshold}"
        for n in other_name_cols
    ]
    ors_string = " OR ".join(ors)
    return f"({ors_string})"


def sql_gen_case_stmt_name_inversion_4(
    col_name: str,
    other_name_cols: list,
    gamma_col_name=None,
    threshold1=1.0,
    threshold2=0.88,
    include_dmeta=False,
):
    """Generate a case expression which can handle name inversions where e.g. surname and forename are inverted

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. surname
        other_name_cols (list): The name of the other columns that contain names e.g. forename1, forename2
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        threshold1 (float, optional): Jaro threshold for almost exact match. Defaults to 1.0.
        threshold2 (float, optional): Jaro threshold for close match Defaults to 0.88.
        include_dmeta (bool, optional): Also allow a dmetaphone match at threshold2

    Returns:
        str: A sql string
    """

    dmeta_statment = ""
    if include_dmeta:
        dmeta_statment = f"""
        when Dmetaphone({col_name}_l) = Dmetaphone({col_name}_r) then 1
        when DmetaphoneAlt({col_name}_l) = DmetaphoneAlt({col_name}_r) then 1
        """

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold1} then 3
    when {_sql_gen_get_or_list_jaro(col_name, other_name_cols, threshold1)} then 2
    {dmeta_statment}
    when jaro_winkler_sim({col_name}_l, {col_name}_r) >= {threshold2} then 1
    else 0 end"""

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


# Array comparison functions


def sql_gen_case_stmt_array_intersect_2(
    col_name: str, gamma_col_name=None, zero_length_is_null=True
):
    """Generate a case comparison which is 1 if the size of the intersection of the arrays is
    one or more.  Otherwise zero.


    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. phone_number
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when size(array_intersect({col_name}_l, {col_name}_r)) >= 1 then 1
    else 0
    end
    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_array_intersect_3(
    col_name: str, gamma_col_name=None, zero_length_is_null=True
):
    """Generate a three level case comparison based on the size of the intersection of the arrays

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. phone_number
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when size(array_intersect({col_name}_l, {col_name}_r)) > 1 then 2
    when size(array_intersect({col_name}_l, {col_name}_r)) = 1 then 1
    else 0
    end
    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def _daexplode(col_name):
    return f"DualArrayExplode({col_name}_l, {col_name}_r)"


def _score_all_pairwise_combinations_distance(col_name, fn_name):

    return f"""
        transform({_daexplode(col_name)},
            x -> {fn_name}(x['_1'], x['_2'] )
            )
    """


def _compare_pairwise_transformed_combinations(col_name, fn_name):

    return f"""
        transform(
            {_daexplode(col_name)},
            x -> {fn_name}(x['_1']) = {fn_name}(x['_2'])
            )
    """


def _compare_pairwise_combinations_leven_prop(col_name):
    return f"""
        transform(
            {_daexplode(col_name)},
            x -> levenshtein(x['_1'], x['_2'])/((length(x['_1']) + length(x['_2']))/2)
            )
    """


def _jaro_winkler_array(col_name):
    return _score_all_pairwise_combinations_distance(col_name, "jaro_winkler_sim")


def _leven_array(col_name):
    return _score_all_pairwise_combinations_distance(col_name, "levenshtein")


def _dmeta_array(col_name):
    # Is at least one true
    return f"""exists(
    {_compare_pairwise_transformed_combinations(col_name, "Dmetaphone")},
    x -> x
    )
    """


def _size_intersect(col_name):
    return f"size(array_intersect({col_name}_l, {col_name}_r))"


def sql_gen_case_stmt_array_combinations_leven_abs_3(
    col_name: str,
    threshold1: int = 0,
    threshold2: int = 2,
    gamma_col_name=None,
    zero_length_is_null=True,
):
    """Compare all combinations of values in input arrays.  Gamma level 2 if minimum levenshtein score is <=
    threshold1.  Gamma level 1 if min score is  <= threshold2.  Otherwise level 0

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. surname
        threshold1 (int, optional):  Defaults to 1.
        threshold2 (int, optional):  Defaults to 2.
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when {_size_intersect(col_name)} >= 1 then 2
    when array_min({_leven_array(col_name)}) <= {threshold1} then 2
    when array_min({_leven_array(col_name)}) <= {threshold2} then 1
    else 0
    end
    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_array_combinations_leven_rel_3(
    col_name: str,
    threshold1: float = 0.0,
    threshold2: float = 0.1,
    gamma_col_name=None,
    zero_length_is_null=True,
):
    """Compare all combinations of values in input arrays.  Gamma level 2 if levenshtein score as
    percentage of average length is <=threshold1.  Gamma level 1 if min score is  <= threshold2.
    Otherwise level 0

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. surname
        threshold1 (int, optional):  Defaults to 0.05 (i.e. 5%)
        threshold2 (int, optional):  Defaults to 0.1 (i.e. 10%)
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when {_size_intersect(col_name)} >= 1 then 2
    when array_min({_compare_pairwise_combinations_leven_prop(col_name)}) <= {threshold1} then 2
    when array_min({_compare_pairwise_combinations_leven_prop(col_name)}) <= {threshold2} then 1
    else 0
    end
    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_array_combinations_jaro_3(
    col_name: str,
    threshold1=1.0,
    threshold2=0.88,
    gamma_col_name=None,
    zero_length_is_null=True,
):
    """Compare all combinations of values in input arrays.  Gamma level 2 if max jaro_winkler score is >=
    threshold1.  Gamma level 1 if max score is  >= threshold2.  Otherwise level 0

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. surname
        threshold1 (int, optional):  Defaults to 1.
        threshold2 (int, optional):  Defaults to 0.88.
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when {_size_intersect(col_name)} >= 1 then 2
    when array_max({_jaro_winkler_array(col_name)}) >= {threshold1} then 2
    when array_max({_jaro_winkler_array(col_name)}) >= {threshold2} then 1
    else 0
    end

    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)


def sql_gen_case_stmt_array_combinations_jaro_dmeta_4(
    col_name: str,
    threshold1=1.0,
    threshold2=0.88,
    gamma_col_name=None,
    zero_length_is_null=True,
):
    """Compare all combinations of values in input arrays.
    Gamma level 3 if max jaro_winkler score is >= threshold1
    Gamma level 2 if there's at least one match on dmetaphone
    Gamma level 1 if max jaro_winkler score is >= threshold2
    else Gamma level 0

    Args:
        col_name (str): The name of the column we want to generate a custom case expression for e.g. surname
        threshold1 (int, optional):  Defaults to 1.0.
        threshold2 (int, optional):  Defaults to 0.88.
        gamma_col_name (str, optional): . The name of the column, for the alias e.g. surname
        zero_length_is_null (bool, optional):  Whether to treat a zero length array as a null. Defaults to True.
    """

    zero_length_expr = ""
    if zero_length_is_null:
        zero_length_expr = (
            f"when size({col_name}_l) = 0 or size({col_name}_r) = 0 then -1"
        )

    c = f"""case
    when {col_name}_l is null or {col_name}_r is null then -1
    {zero_length_expr}
    when {_size_intersect(col_name)} >= 1 then 3
    when array_max({_jaro_winkler_array(col_name)}) >= {threshold1} then 3
    when {_dmeta_array(col_name)} then 2
    when array_max({_jaro_winkler_array(col_name)}) >= {threshold2} then 1
    else 0
    end

    """

    return _add_as_gamma_to_case_statement(c, gamma_col_name)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import lit
from splink.vertically_concat import vertically_concatenate_datasets
from splink.lower_id_on_lhs import lower_id_to_left_hand_side
from splink.blocking import _get_columns_to_retain_blocking, sql_gen_comparison_columns
from splink.settings import Settings
from splink.blocking import block_using_rules
from splink.gammas import add_gammas
from splink.maximisation_step import run_maximisation_step
from splink.model import Model
from splink.cluster import _check_graphframes_installation
from splink.default_settings import complete_settings_dict


def estimate_m_from_labels(
    settings: dict,
    df_or_dfs: DataFrame,
    labels: DataFrame,
    use_connected_components,
    fix_m_probabilities=False,
):
    """Estimate m values from labels

    Args:
        settings (dict): splink settings dictionary
        df_or_dfs (DataFrame): (DataFrame or list of DataFrames, optional):
        labels (DataFrame): Labelled data.
            For link or link and dedupe, should have columns:
            'source_dataset_l', 'unique_id_l', 'source_dataset_r', and 'unique_id_r'
            For dedupe only, only needs 'unique_id_l' and 'unique_id_r' columns
        use_connected_components (bool, optional): Whether to use the connected components approach.
            Defaults to True.  Described here: https://github.com/moj-analytical-services/splink/issues/245
        fix_m_probabilities (bool, optional): If True, output comparison column settings will have
            fix_u_probabilities set to True.  Defaults to False.

    """

    # dfs is a list of dfs irrespective of whether input was a df or list of dfs
    if type(df_or_dfs) == DataFrame:
        dfs = [df_or_dfs]
    else:
        dfs = df_or_dfs

    spark = dfs[0].sql_ctx.sparkSession

    if use_connected_components:
        _check_graphframes_installation(spark)

    df_nodes = vertically_concatenate_datasets(dfs)

    settings_complete = complete_settings_dict(settings, spark)
    if settings_complete["link_type"] == "dedupe_only":
        use_source_dataset = False
    else:
        use_source_dataset = True

    source_dataset_colname = settings_complete["source_dataset_column_name"]
    uid_colname = settings_complete["unique_id_column_name"]

    if use_connected_components:
        df_gammas = _get_comparisons_using_connected_components(
            df_nodes,
            labels,
            settings_complete,
            use_source_dataset,
            source_dataset_colname,
            uid_colname,
        )
    else:
        df_gammas = _get_comparisons_using_joins(
            df_nodes,
            labels,
            settings_complete,
            use_source_dataset,
            source_dataset_colname,
            uid_colname,
        )

    df_e = df_gammas.withColumn("match_probability", lit(1.0))

    model = Model(settings_complete, spark)
    run_maximisation_step(df_e, model, spark)

    settings_with_m_dict = model.current_settings_obj.settings_dict

    # We want to add m probabilities from these estimates to the settings_with_u object
    settings_obj = Settings(settings)

    settings_obj.overwrite_m_u_probs_from_other_settings_dict(
        settings_with_m_dict, overwrite_u=False
    )

    for cc in settings_obj.comparison_columns_list:
        if fix_m_probabilities:
            cc.fix_m_probabilities = True

    return settings_obj.settings_dict


def _get_comparisons_using_connected_components(
    df_nodes,
    df_labels,
    settings_complete,
    use_source_dataset,
    source_dataset_colname,
    uid_colname,
):
    from graphframes import GraphFrame

    spark = df_nodes.sql_ctx.sparkSession

    if use_source_dataset:
        uid_node = f"concat({source_dataset_colname}, '-__-',{uid_colname}) as id"
        uid_r = f"concat({source_dataset_colname}_l, '-__-',{uid_colname}_l) as src"
        uid_l = f"concat({source_dataset_colname}_r, '-__-',{uid_colname}_r) as dst"
    else:
        uid_node = f"{uid_colname} as id"
        uid_r = f"{uid_colname}_l as src"
        uid_l = f"{uid_colname}_r as dst"

    cc_nodes = df_nodes.selectExpr(uid_node)
    edges = df_labels.selectExpr(uid_l, uid_r)
    g = GraphFrame(cc_nodes, edges)
    g = g.dropIsolatedVertices()
    cc = g.connectedComponents()

    df_nodes.createOrReplaceTempView("df_nodes")
    cc.createOrReplaceTempView("cc")

    if use_source_dataset:
        join_col_expr = (
            f"concat(df_nodes.{source_dataset_colname}, '-__-',df_nodes.{uid_colname})"
        )
    else:
        join_col_expr = f"df_nodes.{uid_colname}"

    sql = f"""
    select df_nodes.*, cc.component as cluster
    from df_nodes
    inner join cc
    on cc.id = {join_col_expr}
    """

    df_with_cluster = spark.sql(sql)

    settings_complete["blocking_rules"] = ["l.cluster = r.cluster"]

    df_comparison = block_using_rules(settings_complete, df_with_cluster, spark)
    df_gammas = add_gammas(df_comparison, settings_complete, spark)

    return df_gammas


def _get_comparisons_using_joins(
    df_nodes,
    df_labels,
    settings_complete,
    use_source_dataset,
    source_dataset_colname,
    uid_colname,
):
    spark = df_nodes.sql_ctx.sparkSession
    df_labels = lower_id_to_left_hand_side(
        df_labels, source_dataset_colname, uid_colname
    )

    df_nodes.createOrReplaceTempView("df_nodes")
    df_labels.createOrReplaceTempView("df_labels")

    columns_to_retain = _get_columns_to_retain_blocking(settings_complete, df_nodes)

    sql_select_expr = sql_gen_comparison_columns(columns_to_retain)

    if use_source_dataset:

        sql = f"""
        select {sql_select_expr}, '0' as match_key
        from df_nodes as l
        inner join df_labels
            on l.{source_dataset_colname} = df_labels.{source_dataset_colname}_l and l.{uid_colname} = df_labels.{uid_colname}_l
        inner join df_nodes as r
            on r.{source_dataset_colname} = df_labels.{source_dataset_colname}_r and r.{uid_colname} = df_labels.{uid_colname}_r
        """
    else:
        sql = f"""
        select {sql_select_expr}, '0' as match_key
        from df_nodes as l
        inner join df_labels
            on l.{uid_colname} = df_labels.{uid_colname}_l
        inner join df_nodes as r
            on r.{uid_colname} = df_labels.{uid_colname}_r
        """

    df_comparison = spark.sql(sql)
    df_gammas = add_gammas(df_comparison, settings_complete, spark)
    return df_gammas



import math


def jw_sim_py(
    first,
    second,  # modification from original to not use other imput parameters
):
    """
    Jaro-Winkler similarity calculated exactly as in stringutils.similarity_jaro_winkler in Apache Commons
    using a modified version of the algorithm implemented by 'Jean-Bernard Ratte - jean.bernard.ratte@unary.ca'
    found at https://github.com/nap/jaro-winkler-distance

    used under the Apache License, Version 2.0 , with modifictions explicitly marked
    """

    if not first or not second:
        return 0.0  # modification from original to give 0.0 in case of nulls instead of a an exception

    scaling = 0.1  # modification from original to have constant scaling in order to have a comparable result to the stringutils implementation

    def _get_diff_index(first, second):
        if first == second:
            return -1

        if not first or not second:
            return 0

        max_len = min(len(first), len(second))
        for i in range(0, max_len):
            if not first[i] == second[i]:
                return i

        return max_len

    def _score(first, second):
        shorter, longer = first.lower(), second.lower()

        if len(first) > len(second):
            longer, shorter = shorter, longer

        m1 = _get_matching_characters(shorter, longer)
        m2 = _get_matching_characters(longer, shorter)

        if len(m1) == 0 or len(m2) == 0:
            return 0.0

        return (
            float(len(m1)) / len(shorter)
            + float(len(m2)) / len(longer)
            + float(len(m1) - _transpositions(m1, m2)) / len(m1)
        ) / 3.0

    def _get_prefix(first, second):
        if not first or not second:
            return ""

        index = _get_diff_index(first, second)
        if index == -1:
            return first

        elif index == 0:
            return ""

        else:
            return first[0:index]

    def _transpositions(first, second):
        return math.floor(
            len([(f, s) for f, s in zip(first, second) if not f == s]) / 2.0
        )

    def _get_matching_characters(first, second):
        common = []
        limit = math.floor(min(len(first), len(second)) / 2)

        for i, l in enumerate(first):
            left, right = int(max(0, i - limit)), int(min(i + limit + 1, len(second)))
            if l in second[left:right]:
                common.append(l)
                second = (
                    second[0 : second.index(l)] + "*" + second[second.index(l) + 1 :]
                )

        return "".join(common)

    jaro = _score(first, second)
    cl = min(len(_get_prefix(first, second)), 4)

    return round((jaro + (scaling * cl * (1.0 - jaro))) * 100.0) / 100.0



import pkg_resources
from functools import lru_cache
import math
from jsonschema import validate, ValidationError

import json
import copy

from typeguard import typechecked


@lru_cache()
def _get_schema(setting_dict_should_be_complete=False):
    with pkg_resources.resource_stream(
        __name__, "files/settings_jsonschema.json"
    ) as io:
        schema = json.load(io)

    if not setting_dict_should_be_complete:
        return schema

    if setting_dict_should_be_complete:
        schema2 = copy.deepcopy(schema)
        schema2["required"] = [
            "proportion_of_matches",
            "em_convergence",
            "unique_id_column_name",
            "comparison_columns",
        ]

        schema2["comparison_columns"]["items"]["required"] = [
            "col_name",
            "case_expression",
            "m_probabilities",
            "u_probabilities",
        ]
        return schema2


@typechecked
def validate_settings_against_schema(settings_dict: dict):
    """Validate a splink settings object against its jsonschema

    Args:
        settings_dict (dict): A splink settings dictionary

    Raises:
        ValidationError: [description]

    Returns:
        [type]: [description]
    """

    schema = _get_schema()
    exception_raised = False
    try:
        validate(settings_dict, schema)
    except Exception as e:
        message = (
            "There is an error in your settings dictionary. "
            "To quickly write a valid settings dictionary using autocompelte you might want to try "
            "our online tool https://moj-analytical-services.github.io/splink_settings_editor/ or you can use "
            "the autocomplete features of VS Code - just copy and paste code from the following gist "
            "into VS Code, setting language mode to json, or having saved the file as a .json file\n"
            "https://gist.github.com/RobinL/cfe1152dbd33ae26e05a43d9a0ec85b9"
            "\n\n"
            "The details of the error are as follows:"
            "\n"
        )
        message = message + str(e)
        exception_raised = True

    if exception_raised:
        raise ValidationError(message)


def get_default_value_from_schema(key, is_column_setting):
    schema = _get_schema()

    if is_column_setting:
        return schema["properties"]["comparison_columns"]["items"]["properties"][key][
            "default"
        ]
    else:
        return schema["properties"][key]["default"]


def validate_input_datasets(df, completed_settings_obj):
    """Check that the input datasets contain the columns needed to run the model

    Args:
        df (DataFrame): Dataframe to be used in splink - either the original df, or the concatenation of the input list of dfs
        completed_settings_obj (Settings): Settings object
    """
    cols_present = set(df.columns)

    cols_needed = set()

    s = completed_settings_obj
    cols_needed.add(s["unique_id_column_name"])
    if completed_settings_obj["link_type"] != "dedupe_only":
        cols_needed.add(s["source_dataset_column_name"])

    for cc in s.comparison_columns_list:
        cols_needed.update(cc.input_cols_used)

    required_cols_not_present = cols_needed - cols_present
    if len(required_cols_not_present):
        raise ValueError(
            f"Cannot find columns {required_cols_not_present} "
            "For Splink to be work with the settings provided, "
            "your input dataframes must  include the following columns "
            f"{cols_needed}"
        )


def validate_link_type(df_or_dfs, settings):
    if type(df_or_dfs) == list:
        if "link_type" in settings:
            if settings["link_type"] == "dedupe_only":
                raise ValueError(
                    "If you provide a list of dfs, link_type must be "
                    "link_only or link_and_dedupe, not dedupe_only"
                )


def validate_probabilities(settings_dict):
    from .settings import Settings

    settings_obj = Settings(settings_dict)

    for cc in settings_obj.comparison_columns_list:

        for mu_probabilities in ["m_probabilities", "u_probabilities"]:

            if mu_probabilities in cc.column_dict:
                if None in cc[mu_probabilities]:
                    raise ValueError(
                        f"Your {mu_probabilities} for {cc.name} contain None "
                        "They should all be populated and sum to 1"
                    )

                sum_p = sum(cc[mu_probabilities])

                if not math.isclose(sum_p, 1.0, rel_tol=1e-9, abs_tol=0.0):
                    raise ValueError(
                        f"Your {mu_probabilities} for {cc.name} do not sum to 1 "
                        "They should all be populated and sum to 1"
                    )

                num_levels = cc.num_levels

                if len(cc[mu_probabilities]) != num_levels:
                    raise ValueError(
                        f"Number of probs provided in {mu_probabilities}  in {cc.name} "
                        "is not equal to number of levels specified"
                    )



from copy import deepcopy
from splink.default_settings import _normalise_prob_list

from .blocking import block_using_rules
from .gammas import add_gammas
from .maximisation_step import run_maximisation_step
from .model import Model
from .default_settings import complete_settings_dict
from .vertically_concat import vertically_concatenate_datasets

import warnings

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.functions import lit


def _num_target_rows_to_rows_to_sample(target_rows):
    # Number of rows generated by cartesian product is
    # n(n-1)/2, where n is input rows
    # We want to set a target_rows = t, the number of
    # rows generated by Splink and find out how many input rows
    # we need to generate targer rows
    #     Solve t = n(n-1)/2 for n
    #     https://www.wolframalpha.com/input/?i=Solve%5Bt%3Dn+*+%28n+-+1%29+%2F+2%2C+n%5D
    sample_rows = 0.5 * ((8 * target_rows + 1) ** 0.5 + 1)
    return sample_rows


def estimate_u_values(
    settings: dict,
    df_or_dfs: DataFrame,
    spark: SparkSession,
    target_rows: int = 1e6,
    fix_u_probabilities=False,
):
    """Complete the `u_probabilities` section of the settings object
    by directly estimating `u_probabilities` from raw data (i.e. without
    the expectation maximisation algorithm).

    This procedure takes a sample of the data and generates the cartesian
    product of comparisons.  The validity of the u values rests on the
    assumption that the probability of two comparison in the cartesian
    product being a match is very low.  For large datasets, this is typically
    true.

    Args:
        settings (dict): splink settings dictionary
        df_or_dfs (DataFrame or list of DataFrames, optional):
        spark (SparkSession): SparkSession object
        target_rows (int): The number of rows to generate in the cartesian product.
            If set too high, you can run out of memory.  Default value 1e6. Recommend settings to perhaps 1e7.

    Returns:
        dict: The input splink settings dictionary with the `u_probabilities` completed with
              the estimated values
    """

    # Preserve settings as provided
    orig_settings = deepcopy(settings)

    # Do not modify settings object provided by user either
    settings = deepcopy(settings)

    # For the purpoes of estimating u values, we will not use any blocking
    settings["blocking_rules"] = []

    # Minimise the columns that are needed for the calculation
    settings["retain_matching_columns"] = False
    settings["retain_intermediate_calculation_columns"] = False

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        settings = complete_settings_dict(settings, spark)

    if type(df_or_dfs) == DataFrame:
        dfs = [df_or_dfs]
    else:
        dfs = df_or_dfs

    df = vertically_concatenate_datasets(dfs)

    count_rows = df.count()
    if settings["link_type"] in ["dedupe_only", "link_and_dedupe"]:
        sample_size = _num_target_rows_to_rows_to_sample(target_rows)
        proportion = sample_size / count_rows

    if settings["link_type"] == "link_only":
        sample_size = target_rows ** 0.5
        proportion = sample_size / count_rows

    if proportion >= 1.0:
        proportion = 1.0

    df_s = df.sample(False, proportion)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df_comparison = block_using_rules(settings, df_s, spark)

        df_gammas = add_gammas(df_comparison, settings, spark)

        df_e_product = df_gammas.withColumn("match_probability", lit(0.0))

        model = Model(settings, spark)
        run_maximisation_step(df_e_product, model, spark)
        new_settings = model.current_settings_obj.settings_dict

        for i, col in enumerate(orig_settings["comparison_columns"]):
            u_probs = new_settings["comparison_columns"][i]["u_probabilities"]
            # Ensure non-zero u (https://github.com/moj-analytical-services/splink/issues/161)
            u_probs = [u or 1 / target_rows for u in u_probs]
            u_probs = _normalise_prob_list(u_probs)
            col["u_probabilities"] = u_probs
            if fix_u_probabilities:
                col["fix_u_probabilities"] = True

    return orig_settings



from functools import reduce
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import lit
from typing import List
from typeguard import typechecked


@typechecked
def vertically_concatenate_datasets(dfs: List[DataFrame]):

    # We require datasets strictly to have the same columns
    # And the same types
    cols = set(dfs[0].columns)
    types = set(dfs[0].dtypes)
    for df in dfs:
        these_cols = set(df.columns)
        if these_cols != cols:
            raise ValueError(
                "Each dataframe in the list of dataframes provided to Splink"
                "must have strictly the same columns.  You provided one with "
                f"{these_cols} and another with {cols}"
            )
        these_types = set(df.dtypes)
        if these_types != types:
            raise ValueError(
                "Each dataframe in the list of dataframes provided to Splink"
                "must have strictly the same columns with same data types. "
                " You provided one with "
                f"{these_types} and another with {types}"
            )

    # UnionAll concats by position not name
    col_order = dfs[0].columns
    dfs = [df.select(col_order) for df in dfs]

    df = reduce(DataFrame.unionAll, dfs)

    return df



from copy import deepcopy
from math import log2
import re

from .validate import get_default_value_from_schema

from .charts import load_chart_definition, altair_if_installed_else_json


class ComparisonLevel:
    def __init__(self, level_dict, comparison_column_obj=None, settings_obj=None):
        self.level_dict = level_dict
        self.comparison_column = comparison_column_obj
        self.settings = settings_obj

    def __getitem__(self, i):
        return self.level_dict[i]

    @property
    def is_null(self):
        vector_value = self.comparison_vector_value
        if vector_value:
            if str(vector_value) == "-1":
                return True
            else:
                return False
        else:
            return None

    @property
    def comparison_vector_value(self):
        sql_expr = self.level_dict.get("sql_expr")

        if sql_expr:
            # sql_expr is either in the form 'when ... then x' or 'else y'
            # we want to extract the value of x or y
            # sqlglot needs a full sql case expression, it can't parse e.g. 'else 0'

            # capture value after else using regex
            sql_expr = sql_expr.strip()
            if sql_expr.lower().startswith("else"):
                return sql_expr[4:].strip()

            # https://stackoverflow.com/a/33233868/1779128
            case_value = re.search(r"(?s:.*)then (.*)$", sql_expr, re.IGNORECASE).group(
                1
            )
            if case_value:
                return case_value.strip()
        else:
            return None

    @property
    def not_null(self):
        if self.is_null is None:
            return None
        return not self.is_null

    @property
    def has_m_u(self):

        if self.is_null:
            return True

        m = self.level_dict.get("m_probability")
        u = self.level_dict.get("u_probability")

        if m and u:
            return True

        return False

    @property
    def m(self):
        if self.is_null:
            return 1

        if "m_probability" in self.level_dict:
            return self.level_dict["m_probability"]
        return None

    @property
    def u(self):
        if self.is_null:
            return 1

        if "u_probability" in self.level_dict:
            return self.level_dict["u_probability"]
        return None

    @property
    def bayes_factor(self):
        if self.has_m_u and self.u != 0:
            return self.m / self.u
        else:
            return None

    @property
    def log2_bayes_factor(self):
        bf = self.bayes_factor
        if bf and self.m != 0:
            return log2(bf)

    @property
    def gamma_index(self):
        if "gamma_index" in self.level_dict:
            return self.level_dict["gamma_index"]
        else:
            return None

    @property
    def proportion_of_nonnull_records_in_level(self):

        if self.settings:
            sd = self.settings.settings_dict
            lam = sd.get("proportion_of_matches")

        if lam and self.has_m_u:
            return self.m * lam + self.u * (1 - lam)
        else:
            return None

    def as_dict(self):

        d = {}
        d["label"] = self["label"]

        d["sql_expr"] = self["sql_expr"]

        d["gamma_column_name"] = f"gamma_{self.comparison_column.name}"
        d["level_name"] = self["label"]

        d["gamma_index"] = self.gamma_index
        d["comparison_vector_value"] = self.comparison_vector_value
        d["column_name"] = self.comparison_column.name
        d["num_levels"] = self.comparison_column.num_levels

        d["m_probability"] = self.m
        d["u_probability"] = self.u
        d["bayes_factor"] = self.bayes_factor
        d["log2_bayes_factor"] = self.log2_bayes_factor

        d["level_proportion"] = self.proportion_of_nonnull_records_in_level
        d["is_null"] = self.is_null

        return d


class ComparisonColumn:
    def __init__(self, column_dict, settings_obj=None):
        self.column_dict = column_dict
        self.settings_obj = settings_obj

    def __getitem__(self, i):
        return self.column_dict[i]

    def _dict_key_else_default_value(self, key):
        cd = self.column_dict

        if key in cd:
            return cd[key]
        else:
            return get_default_value_from_schema(key, True)

    @property
    def custom_comparison(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return True
        elif "col_name" in cd:
            return False

    @property
    def columns_used(self):
        cd = self.column_dict
        if "col_name" in cd:
            return [cd["col_name"]]
        elif "custom_name" in cd:
            return cd["custom_columns_used"]

    @property
    def name(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return cd["custom_name"]
        elif "col_name" in cd:
            return cd["col_name"]

    @property
    def has_case_expression_or_comparison_levels(self):
        cd = self.column_dict
        if "case_expression" in cd:
            return True
        elif "comparison_levels" in cd:
            return True
        else:
            return False

    @property
    def gamma_name(self):
        return f"gamma_{self.name}"

    @property
    def num_levels(self):
        cd = self.column_dict

        cld = self.comparison_levels_dict
        if cld:
            comparison_levels_keys = cld.keys()

            comparison_levels_keys = [
                k for k in comparison_levels_keys if str(k) != "-1"
            ]
            return len(comparison_levels_keys)
        if "num_levels" in cd:
            return cd["num_levels"]

    @property
    def input_cols_used(self):
        cd = self.column_dict
        if "custom_name" in cd:
            return cd["custom_columns_used"]
        elif "col_name" in cd:
            return [cd["col_name"]]

    def _attach_m_u_to_comparison_levels(self, comparison_levels):

        m_probabilities = self.column_dict.get("m_probabilities")
        if m_probabilities:

            for level in comparison_levels:
                cl = ComparisonLevel(level)
                m_index = int(cl.comparison_vector_value)
                if cl.not_null:
                    level["m_probability"] = m_probabilities[m_index]

        u_probabilities = self.column_dict.get("u_probabilities")
        if u_probabilities:
            counter = 0
            for level in comparison_levels:
                cl = ComparisonLevel(level)
                u_index = int(cl.comparison_vector_value)
                if cl.not_null:
                    level["u_probability"] = u_probabilities[u_index]

        return comparison_levels

    @property
    def comparison_levels_dict(self):
        cd = self.column_dict

        if "comparison_levels" not in cd:
            return None

        comparison_levels = deepcopy(cd["comparison_levels"])

        results = {}
        for level in comparison_levels:
            cl = ComparisonLevel(level, self, self.settings_obj)
            key = cl.comparison_vector_value
            results[key] = cl
            cl.level_dict["gamma_index"] = int(key)

        comparison_levels = self._attach_m_u_to_comparison_levels(comparison_levels)
        return results

    @property
    def comparison_levels_list(self):
        return list(self.comparison_levels_dict.values())

    @property
    def term_frequency_adjustments(self):
        cd = self.column_dict
        return cd["term_frequency_adjustments"]

    def df_e_row_intuition_dict(self, row_dict):

        gamma_value = str(int(row_dict[self.gamma_name]))
        cl = self.comparison_levels_dict[gamma_value]
        row_desc = cl.as_dict()

        row_desc["value_l"] = ", ".join(
            [str(row_dict[c + "_l"]) for c in self.columns_used]
        )
        row_desc["value_r"] = ", ".join(
            [str(row_dict[c + "_r"]) for c in self.columns_used]
        )

        return row_desc

    def set_m_probability(self, level: int, prob: float, force: bool = False):
        cd = self.column_dict
        fixed = self._dict_key_else_default_value("fix_m_probabilities")
        if not fixed or force:
            cd["m_probabilities"][level] = prob

    def set_u_probability(self, level: int, prob: float, force: bool = False):
        cd = self.column_dict
        fixed = self._dict_key_else_default_value("fix_u_probabilities")
        if not fixed or force:
            cd["u_probabilities"][level] = prob

    def reset_probabilities(self, force: bool = False):
        cd = self.column_dict
        fixed_m = self._dict_key_else_default_value("fix_m_probabilities")
        fixed_u = self._dict_key_else_default_value("fix_u_probabilities")
        if not fixed_m or force:
            if "m_probabilities" in cd:
                cd["m_probabilities"] = [None for c in cd["m_probabilities"]]

        if not fixed_u or force:
            if "u_probabilities" in cd:
                cd["u_probabilities"] = [None for c in cd["u_probabilities"]]

    def as_rows(self, proportion_of_matches=None):
        """Convert to rows e.g. to use to plot
        in a chart"""
        rows = []
        # This should iterate over the `comparison_levels` dict
        for comparison_level in self.comparison_levels_list:
            r = comparison_level.as_dict()
            rows.append(r)
        return rows

    def _repr_pretty_(self, p, cycle):

        p.text("------------------------------------")
        p.break_()
        p.text(f"Comparison of {self.name}")
        p.break_()
        for row in self.as_rows():
            p.text(f"{row['level_name']}")
            p.break_()
            value_m = (
                f"{row['m_probability']*100:.3g}%" if row["m_probability"] else "None"
            )
            p.text(f"   Proportion in level amongst matches:       {value_m}")
            p.break_()
            value_u = (
                f"{row['u_probability']*100:.3g}%" if row["u_probability"] else "None"
            )
            p.text(f"   Proportion in level amongst non-matches:   {value_u}%")
            p.break_()
            value_b = (
                f"{row['bayes_factor']*100:.3g}%" if row["bayes_factor"] else "None"
            )
            p.text(f"   Bayes factor:                              {value_b}")
            p.break_()


class Settings:
    def __init__(self, settings_dict):
        self.settings_dict = deepcopy(settings_dict)

    def __getitem__(self, i):
        return self.settings_dict[i]

    def __setitem__(self, key, value):
        self.settings_dict[key] = value

    def complete_settings_dict(self, spark):
        """Complete all fields in the setting dictionary
        taking values from defaults"""
        from .default_settings import complete_settings_dict

        self.settings_dict = complete_settings_dict(self.settings_dict, spark)

    @property
    def comparison_column_dict(self):
        sd = self.settings_dict
        lookup = {}
        for c in sd["comparison_columns"]:

            cc = ComparisonColumn(c, self)
            name = cc.name
            lookup[name] = cc
        return lookup

    @property
    def comparison_columns_list(self):
        return list(self.comparison_column_dict.values())

    @property
    def any_cols_have_tf_adjustments(self):
        return any([c.term_frequency_adjustments for c in self.comparison_columns_list])

    def get_comparison_column(self, col_name_or_custom_name):
        if col_name_or_custom_name in self.comparison_column_dict:
            return self.comparison_column_dict[col_name_or_custom_name]
        else:
            raise KeyError(
                f"You requested comparison column {col_name_or_custom_name}"
                " but it does not exist in the settings object"
            )

    def reset_all_probabilities(self, force: bool = False):
        sd = self.settings_dict
        sd["proportion_of_matches"] = None
        for c in self.comparison_columns_list:
            c.reset_probabilities(force=force)

    def m_u_as_rows(self, drop_null=True):
        """Convert to rows e.g. to use to plot
        in a chart"""
        rows = []
        for cc in self.comparison_columns_list:
            rows.extend(cc.as_rows())
        if drop_null:
            rows = [r for r in rows if not r["is_null"]]
        return rows

    def set_m_probability(
        self, name: str, level: int, prob: float, force: bool = False
    ):
        cc = self.get_comparison_column(name)
        cc.set_m_probability(level, prob, force)

    def set_u_probability(
        self, name: str, level: int, prob: float, force: bool = False
    ):
        cc = self.get_comparison_column(name)
        cc.set_u_probability(level, prob, force)

    def overwrite_m_u_probs_from_other_settings_dict(
        self, incoming_settings_dict, overwrite_m=True, overwrite_u=True
    ):
        """Overwrite the m and u probabilities with
        the values from another settings dict where comparison column exists
        in this settings object
        """
        incoming_settings_obj = Settings(incoming_settings_dict)

        for cc_incoming in incoming_settings_obj.comparison_columns_list:
            if cc_incoming.name in self.comparison_column_dict:
                cc_existing = self.get_comparison_column(cc_incoming.name)

                if overwrite_m:
                    if "m_probabilities" in cc_incoming.column_dict:
                        cc_existing.column_dict["m_probabilities"] = cc_incoming[
                            "m_probabilities"
                        ]

                if overwrite_u:
                    if "u_probabilities" in cc_incoming.column_dict:
                        cc_existing.column_dict["u_probabilities"] = cc_incoming[
                            "u_probabilities"
                        ]

    def remove_comparison_column(self, name):
        removed = False
        new_ccs = []
        for cc in self.comparison_columns_list:
            if cc.name != name:
                new_ccs.append(cc.column_dict)
            else:
                removed = True
        self.settings_dict["comparison_columns"] = new_ccs
        if not removed:
            raise ValueError(f"Could not find a column named {name}")

    def probability_distribution_chart(self):  # pragma: no cover
        chart_path = "probability_distribution_chart.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.m_u_as_rows()
        chart["title"]["subtitle"] += f" {self['proportion_of_matches']:.3g}"

        return altair_if_installed_else_json(chart)

    def bayes_factor_chart(self):  # pragma: no cover
        chart_path = "bayes_factor_chart_def.json"
        chart = load_chart_definition(chart_path)
        rows = self.m_u_as_rows()

        chart["data"]["values"] = rows
        return altair_if_installed_else_json(chart)

    def _repr_pretty_(self, p, cycle):  # pragma: no cover
        if cycle:
            p.text("ComparisonColumn()")
        else:
            value = (
                f"{self['proportion_of_matches']:.4g}"
                if self["proportion_of_matches"]
                else "None"
            )
            p.text(f"λ (proportion of matches) = {value}")
            for c in self.comparison_columns_list:
                p.break_()
                p.pretty(c)



from statistics import harmonic_mean
from copy import deepcopy

from splink.charts import load_chart_definition, altair_if_installed_else_json
from splink.settings import Settings
import warnings


def _apply_aggregate_function(zipped_probs, aggregate_function):
    result = []
    for probs_list in zipped_probs:
        try:
            reduced = aggregate_function(probs_list)
        except Exception as e:
            warnings.warn(
                "The aggregation function produced an error when "
                f"operating on the following data: {probs_list}. "
                "The result of this aggreation has been set to None. "
                f"The error was {e}"
            )
            reduced = None
        result.append(reduced)
    return result


def _format_probs_for_report(probs):
    try:
        probs_as_strings = [f"{p:.4g}" for p in probs]
    except TypeError:
        probs_as_strings = [f"{p}" for p in probs]
    return f"{probs_as_strings}"


def _filter_nones(list_of_lists):
    def filter_none(sublist):
        return [item for item in sublist if item is not None]

    return [filter_none(sublist) for sublist in list_of_lists]


def _zip_m_and_u_probabilities(cc_estimates: list):
    """Groups together the different estimates of the same parameter.

    e.g. turns:
    [{"m_probabilities":[ma0,ma1],{"m_probabilities":[ua0,ua1],...},
        {"m_probabilities":[mb0,mb1],{"m_probabilities":[ub0,ub1],...}]

    into

    {"zipped_m": [(ma0,mb0), (ma1, mb1)],
        "zipped_u": [(ua0,ub0), (ua1, ub1)]}
    """

    zipped_m_probs = zip(*[cc["m_probabilities"] for cc in cc_estimates])
    zipped_m_probs = _filter_nones(zipped_m_probs)
    zipped_u_probs = zip(*[cc["u_probabilities"] for cc in cc_estimates])
    zipped_u_probs = _filter_nones(zipped_u_probs)

    return {"zipped_m": zipped_m_probs, "zipped_u": zipped_u_probs}


def combine_cc_estimates(cc_estimates: list, aggregate_function=harmonic_mean):
    """cc_estimates is a list of the different estaimtes for a single comparison column
    e.g. all of the different comparison columns for forename from params_list
    """

    zipped = _zip_m_and_u_probabilities(cc_estimates)

    m_probs = _apply_aggregate_function(zipped["zipped_m"], aggregate_function)
    u_probs = _apply_aggregate_function(zipped["zipped_u"], aggregate_function)

    cc = deepcopy(cc_estimates[0].column_dict)
    cc["m_probabilities"] = m_probs
    cc["u_probabilities"] = u_probs
    return cc


class ModelCombiner:
    def __init__(self, input_dicts: list):

        self.input_dicts = input_dicts
        self.named_settings_dict = {}
        for i in input_dicts:
            name = i["name"]
            settings = i["model"].current_settings_obj
            self.named_settings_dict[name] = settings

        self.settings_obj_list = list(self.named_settings_dict.values())

    def _groups_of_comparison_columns_by_name(self):
        """
        The user inputs a list of parameter estimates, each of which
        contains a settings dict

        If the input data is:
        Model list element 1:  Estimate name 'forename blocking'
            Comparison columns: [surname, dob, email]
        Model list element 2:  Estimate name 'surname blocking'
            Comparison columns: [forename, dob, email]
        Model list element 3:  Estimate name 'dob blocking'
            Comparison columns: [forename, surname, email]

        We want to group by comparison column name, respecting the fact
        that not all models have all comparison columns

        This function gives you back a dict in the form:
        {
            "forename": {"surname blocking": cc, "dob_blocking": cc},
            "surname": [etc]
            "dob": [etc]
            "email": [etc]

        }
        """
        combined_cc = {}
        # For each model which has been estimated
        for estimate_name, settings_estimate in self.named_settings_dict.items():
            # For each comparison column in this model
            for cc_name, cc in settings_estimate.comparison_column_dict.items():
                # Create or add to dict which contains the different estimate
                # for this column, using estimate names as keys
                if cc_name not in combined_cc:
                    combined_cc[cc_name] = {}
                combined_cc[cc_name][estimate_name] = cc
        return combined_cc

    def get_combined_settings_dict(self, aggregate_function=harmonic_mean):

        new_settings = deepcopy(self.settings_obj_list[0].settings_dict)

        new_comparison_columns = []
        gathered = self._groups_of_comparison_columns_by_name()

        # For each comparison column (first name, surname etc)
        for dict_of_ccs in gathered.values():
            ccs = list(dict_of_ccs.values())
            # Take the average of each parameter estimates
            combined = combine_cc_estimates(ccs, aggregate_function)
            new_comparison_columns.append(combined)

        new_settings["comparison_columns"] = new_comparison_columns

        new_blocking_rules = []
        for settings_dict in self.settings_obj_list:
            new_blocking_rules.extend(settings_dict["blocking_rules"])

        new_settings["blocking_rules"] = new_blocking_rules

        global_lambdas = []
        for input_dict in self.input_dicts:
            # If key exists, use in computation of global lambda
            # else ignore
            if "comparison_columns_for_global_lambda" in input_dict:
                gl = self._estimate_global_lambda_from_blocking_specific_lambda(
                    input_dict
                )
                global_lambdas.append(gl)
        if len(global_lambdas) > 0:
            global_lambda = aggregate_function(global_lambdas)
        else:
            global_lambda = None

        new_settings["proportion_of_matches"] = global_lambda

        return new_settings

    def _estimate_global_lambda_from_blocking_specific_lambda(self, model_dict):

        bayes_factor = 1
        for cc in model_dict["comparison_columns_for_global_lambda"]:
            m = cc["m_probabilities"][-1]
            u = cc["u_probabilities"][-1]
            b = m / u
            bayes_factor = bayes_factor * b

        # https://observablehq.com/@robinl/conditional-independence-and-repeated-application-of-bay
        # https://www.wolframalpha.com/input/?i=solve%5Bm%3Db*%CE%BB%2F%28b*%CE%BB+%2B+%281-%CE%BB%29%29%5D
        blocking_specific_lambda = model_dict["model"].current_settings_obj[
            "proportion_of_matches"
        ]

        global_lambda = blocking_specific_lambda / (
            bayes_factor
            + blocking_specific_lambda
            - bayes_factor * blocking_specific_lambda
        )

        return global_lambda

    def summary_report(self, aggregate_function=harmonic_mean, summary_name="combined"):

        lines = []
        gathered = self._groups_of_comparison_columns_by_name()

        combined_settings = self.get_combined_settings_dict(
            aggregate_function=aggregate_function
        )
        combined_settings = Settings(combined_settings)

        for cc_name, dict_of_ccs in gathered.items():

            lines.append(f"Column name: {cc_name}")

            lines.append(f"    m probabilities")
            for estimate_name, cc in dict_of_ccs.items():
                m_probs = _format_probs_for_report(cc["m_probabilities"])
                lines.append(f"        {estimate_name:<15}: {m_probs}")

            cc = combined_settings.get_comparison_column(cc_name)
            m_probs = _format_probs_for_report(cc["m_probabilities"])
            summary = f"{summary_name} value:"
            lines.append(f"        {summary:<15}: {m_probs}")

            lines.append(f"    u probabilities")

            for estimate_name, cc in dict_of_ccs.items():
                m_probs = _format_probs_for_report(cc["u_probabilities"])
                lines.append(f"        {estimate_name:<15}: {m_probs}")

            cc = combined_settings.get_comparison_column(cc_name)
            m_probs = _format_probs_for_report(cc["u_probabilities"])
            summary = f"{summary_name} value:"
            lines.append(f"        {summary:<15}: {m_probs}")

        return "\n".join(lines)

    def _estimates_as_rows(self):
        """A list of dicts represeting
        all the param estimates which can be passed
        t"""
        rows = []

        gathered = self._groups_of_comparison_columns_by_name()
        for dict_of_ccs in gathered.values():
            for estimate_name, cc in dict_of_ccs.items():
                new_rows = cc.as_rows()
                for r in new_rows:
                    r["estimate_name"] = estimate_name
                rows.extend(new_rows)

        return rows

    def comparison_chart(self):
        chart_def = load_chart_definition("compare_estimates.json")
        chart_def["data"]["values"] = self._estimates_as_rows()

        return altair_if_installed_else_json(chart_def)

    def _repr_pretty_(self, p, cycle):
        p.text(
            self.summary_report(
                summary_name="harmonic_mean", aggregate_function=harmonic_mean
            )
        )



from typing import Callable, Union, List
from typeguard import typechecked

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from splink.validate import (
    validate_settings_against_schema,
    validate_input_datasets,
    validate_link_type,
    validate_probabilities,
)
from splink.model import Model, load_model_from_json
from splink.case_statements import _check_jaro_registered
from splink.blocking import block_using_rules
from splink.gammas import add_gammas
from splink.iterate import iterate
from splink.expectation_step import run_expectation_step
from splink.term_frequencies import add_term_frequencies
from splink.vertically_concat import (
    vertically_concatenate_datasets,
)
from splink.break_lineage import default_break_lineage_blocked_comparisons

from splink.default_settings import normalise_probabilities


@typechecked
class Splink:
    def __init__(
        self,
        settings: dict,
        df_or_dfs: Union[DataFrame, List[DataFrame]],
        spark: SparkSession,
        save_state_fn: Callable = None,
        break_lineage_blocked_comparisons: Callable = default_break_lineage_blocked_comparisons,
    ):
        """Splink data linker

        Provides easy access to the core user-facing functionality of splink

        Args:
            settings (dict): splink settings dictionary
            df_or_dfs (Union[DataFrame, List[DataFrame]]): Either a single Spark dataframe to dedupe, or a list of
                Spark dataframes to link and or dedupe. Where `link_type` is `dedupe_only`, should be a single dataframe
                to dedupe. Where `link_type` is `link_only` or `link_and_dedupe`, show be a list of dfs.  Requires
                conformant dataframes (i.e. they must have same columns)
            spark (SparkSession): SparkSession object
            save_state_fn (function, optional):  A function provided by the user that takes one arguments, model (i.e.
                a Model from splink.model), and is executed each iteration.  This is a hook that allows the user to save
                the state between iterations, which is mostly useful for very large jobs which may need to be restarted
                from where they left off if they fail.
            break_lineage_blocked_comparisons (function, optional): Large jobs will likely run into memory errors
                unless the lineage is broken after blocking.  This is a user-provided function that takes one argument
                - df - and allows the user to break lineage.  For example, the function might save df to the AWS s3
                file system, and then reload it from the saved files.

        """

        self.spark = spark
        self.break_lineage_blocked_comparisons = break_lineage_blocked_comparisons

        _check_jaro_registered(spark)

        validate_settings_against_schema(settings)
        validate_link_type(df_or_dfs, settings)

        self.model = Model(settings, spark)
        self.settings_dict = self.model.current_settings_obj.settings_dict
        self.settings_dict = normalise_probabilities(self.settings_dict)
        validate_probabilities(self.settings_dict)
        # dfs is a list of dfs irrespective of whether input was a df or list of dfs
        if type(df_or_dfs) == DataFrame:
            dfs = [df_or_dfs]
        else:
            dfs = df_or_dfs

        self.df = vertically_concatenate_datasets(dfs)
        self.df = add_term_frequencies(self.df, self.model, self.spark)
        validate_input_datasets(self.df, self.model.current_settings_obj)
        self.save_state_fn = save_state_fn

    def manually_apply_fellegi_sunter_weights(self):
        """Compute match probabilities from m and u probabilities specified in the splink settings object

        Returns:
            DataFrame: A spark dataframe including a match probability column
        """
        df_comparison = block_using_rules(self.settings_dict, self.df, self.spark)
        df_gammas = add_gammas(df_comparison, self.settings_dict, self.spark)
        # see https://github.com/moj-analytical-services/splink/issues/187
        df_gammas = self.break_lineage_blocked_comparisons(df_gammas, self.spark)
        return run_expectation_step(df_gammas, self.model, self.spark)

    def get_scored_comparisons(self):
        """Use the EM algorithm to estimate model parameters and return match probabilities.

        Returns:
            DataFrame: A spark dataframe including a match probability column
        """

        df_comparison = block_using_rules(self.settings_dict, self.df, self.spark)

        df_gammas = add_gammas(df_comparison, self.settings_dict, self.spark)

        df_gammas = self.break_lineage_blocked_comparisons(df_gammas, self.spark)

        df_e = iterate(
            df_gammas,
            self.model,
            self.spark,
            save_state_fn=self.save_state_fn,
        )

        # In case the user's break lineage function has persisted it
        df_gammas.unpersist()

        return df_e

    def save_model_as_json(self, path: str, overwrite=False):
        """Save model (settings, parameters and parameter history) as a json file so it can later be re-loaded
        using load_from_json

        Args:
            path (str): Path to the json file.
            overwrite (bool): Whether to overwrite the file if it exsits
        """
        self.model.save_model_to_json_file(path, overwrite=overwrite)


@typechecked
def load_from_json(
    path: str,
    df_or_dfs: Union[DataFrame, List[DataFrame]],
    spark: SparkSession,
    save_state_fn: Callable = None,
):
    """Load a splink model from a json file which has previously been created using 'save_model_as_json'

    Args:
        path (string): path to json file created using Splink.save_model_as_json
        spark (SparkSession): SparkSession object
        df (DataFrame, optional): The dataframe to dedupe. Where `link_type` is `dedupe_only`, the dataframe to dedupe.
            Should be ommitted `link_type` is `link_only` or `link_and_dedupe`.
        save_state_fn (function, optional):  A function provided by the user that takes one argument, params, and is
            executed each iteration.  This is a hook that allows the user to save the state between iterations, which
            is mostly useful for very large jobs which may need to be restarted from where they left off if they fail.

    Returns:
        splink.model.Model
    """
    model = load_model_from_json(path)

    linker = Splink(
        model.current_settings_obj.settings_dict, df_or_dfs, spark, save_state_fn
    )
    linker.model = model
    return linker



import sqlglot
from sqlglot.errors import ParseError
from sqlglot.expressions import Case, Column, Alias, Bracket
import re


def get_columns_used_from_sql(sql):
    column_names = set()
    syntax_tree = sqlglot.parse_one(sql, read="spark")
    for tup in syntax_tree.walk():
        subtree = tup[0]
        if type(subtree) is Column:
            column_names.add(subtree.sql())
        elif type(subtree) is Bracket:
            column_names.add(subtree.text("this"))

    return list(column_names)


def get_columns_used_from_sql_without_l_r_suffix(sql):
    cols = get_columns_used_from_sql(sql)
    cols = [re.sub(r"_L$|_R$", "", c, flags=re.IGNORECASE) for c in cols]
    return list(set(cols))


def _tree_is_alias(syntax_tree):
    return type(syntax_tree) is Alias


def _tree_is_case(syntax_tree):
    return type(syntax_tree) is Case


def _get_top_level_case(sql):
    try:
        syntax_tree = sqlglot.parse_one(sql, read="spark")
    except ParseError as e:
        raise ValueError(f"Error parsing case statement:\n{sql}") from e

    if _tree_is_alias(syntax_tree):
        case = syntax_tree.find(Case)
        if case.depth == 1:
            sql = case.sql()
            case_tree = sqlglot.parse_one(sql, read="spark")
            return case_tree
        else:
            raise ValueError(
                "Error parsing case statement - no case statement found at top level\n"
                f"Statement was: {sql}"
            )
    elif _tree_is_case(syntax_tree):
        return syntax_tree
    else:
        raise ValueError(
            "Error parsing case statement - no case statement found at top level\n"
            f"Statement was: {sql}"
        )


def _parse_top_level_case_statement_from_sql(top_level_case_tree):

    parsed_case_expr = []

    ifs = top_level_case_tree.args["ifs"]
    for i in ifs:
        lit = i.args["true"].sql()

        sql = i.args["this"].sql(dialect="spark")
        sql = f"WHEN {sql} THEN {lit}".format(sql=sql, lit=lit)

        parsed_case_expr.append({"sql_expr": sql, "label": f"level_{lit}"})

    if top_level_case_tree.args.get("default") is not None:
        lit = top_level_case_tree.args.get("default").sql("spark", pretty=True)
        sql = f"ELSE {lit}"
        parsed_case_expr.append({"sql_expr": sql, "label": f"level_{lit}"})

    return parsed_case_expr


def parse_case_statement(sql):

    tree = _get_top_level_case(sql)
    return _parse_top_level_case_statement_from_sql(tree)


def generate_sql_from_parsed_case_expr(parsed_case_expr, col_name=None):
    sql = "CASE\n"
    for value in parsed_case_expr:
        sql_expr = value["sql_expr"]
        sql += f"    {sql_expr}\n"

    sql += "END"
    if col_name:
        sql += f" AS gamma_{col_name}"
    return sql



from collections import OrderedDict


class OrderedSet:
    def __init__(self):
        self._od = OrderedDict()

    def add(self, value):
        self._od[value] = None

    def as_list(self):
        return list(self._od.keys())

    def __iter__(self):
        for key in self._od.keys():
            yield key



import re
from copy import deepcopy

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from functools import reduce


altair_installed = True
try:
    import altair as alt
    from altair import Chart
except ImportError:
    altair_installed = False

_outer_chart_spec_freq = {
    "config": {"view": {"continuousWidth": 400, "continuousHeight": 300}},
    "vconcat": [],
    "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
}

_inner_chart_spec_freq = {
    "hconcat": [
        {
            "data": {"values": None},
            "mark": {"type": "line", "interpolate": "step-after"},
            "encoding": {
                "x": {
                    "type": "quantitative",
                    "field": "percentile_ex_nulls",
                    "sort": "descending",
                    "title": "Percentile",
                },
                "y": {
                    "type": "quantitative",
                    "field": "value_count",
                    "title": "Count of values",
                },
                "tooltip": [
                    {"field": "value_count", "type": "quantitative"},
                    {"field": "percentile_ex_nulls", "type": "quantitative"},
                    {"field": "percentile_inc_nulls", "type": "quantitative"},
                    {"field": "total_non_null_rows", "type": "quantitative"},
                    {"field": "total_rows_inc_nulls", "type": "quantitative"},
                    {"field": "proportion_of_non_null_rows", "type": "quantitative"},
                ],
            },
            "title": {
                "text": "Distribution of counts of values in column",
                "subtitle": "Subtitle Text",
            },
        },
        {
            "data": {"values": None},
            "mark": "bar",
            "encoding": {
                "x": {
                    "type": "nominal",
                    "field": "value",
                    "sort": "-y",
                    "title": None,
                },
                "y": {
                    "type": "quantitative",
                    "field": "value_count",
                    "title": "Value count",
                },
                "tooltip": [
                    {"field": "value", "type": "nominal"},
                    {"field": "value_count", "type": "quantitative"},
                    {"field": "total_non_null_rows", "type": "quantitative"},
                    {"field": "total_rows_inc_nulls", "type": "quantitative"},
                ],
            },
            "title": "Top 20 values by value count",
        },
        {
            "data": {"values": None},
            "mark": "bar",
            "encoding": {
                "x": {
                    "type": "nominal",
                    "field": "value",
                    "sort": "-y",
                    "title": None,
                },
                "y": {
                    "type": "quantitative",
                    "field": "value_count",
                    "title": "Value count",
                },
                "tooltip": [
                    {"field": "value", "type": "nominal"},
                    {"field": "value_count", "type": "quantitative"},
                    {"field": "total_non_null_rows", "type": "quantitative"},
                    {"field": "total_rows_inc_nulls", "type": "quantitative"},
                ],
            },
            "title": "Bottom 20 values by value count",
        },
    ]
}


def _get_inner_chart_spec_freq(percentile_data, top_n_data, bottom_n_data, col_name):

    inner_spec = deepcopy(_inner_chart_spec_freq)

    total_rows_inc_nulls = percentile_data[0]["total_rows_inc_nulls"]
    total_non_null_rows = percentile_data[0]["total_non_null_rows"]
    distinct_value_count = percentile_data[0]["distinct_value_count"]
    perc = total_non_null_rows / total_rows_inc_nulls

    sub = (
        f"In this col, {total_rows_inc_nulls*(1-perc):,.0f} values ({1-perc:,.1%}) are null and there are "
        f"{distinct_value_count} distinct values"
    )
    sub = sub.format(**percentile_data[0])
    inner_spec["hconcat"][0]["data"]["values"] = percentile_data
    inner_spec["hconcat"][0]["title"][
        "text"
    ] = f"Distribution of counts of values in column {col_name}"

    inner_spec["hconcat"][0]["title"]["subtitle"] = sub

    inner_spec["hconcat"][1]["data"]["values"] = top_n_data
    inner_spec["hconcat"][1]["title"] = f"Top {len(top_n_data)} values by value count"

    inner_spec["hconcat"][2]["data"]["values"] = bottom_n_data
    inner_spec["hconcat"][2][
        "title"
    ] = f"Bottom {len(bottom_n_data)} values by value count"

    max_val = top_n_data[0]["value_count"]
    inner_spec["hconcat"][2]["encoding"]["y"]["scale"] = {"domain": [0, max_val]}

    return inner_spec


def _group_name(cols_or_exprs):
    group_name = "_".join(cols_or_exprs)
    group_name = re.sub(r"[^0-9a-zA-Z_\(\),]", " ", group_name)
    group_name = re.sub(r"\s+", " ", group_name)
    return group_name


def _non_array_group(cols_or_exprs, table_name):
    """
    Generate a sql expression that will yield a table with value counts grouped by the provided
    column expression e.g. "dmetaphone(name)" or combination of column expression
    (e.g. ["dmetaphone(first_name)" , "dmetaphone(surname)"]..

    Where an array is provided, value counts will be grouped by the concatenation of
    all the column expressions


    """
    # If user has provided a string, make it into a list of length one for convenience
    if type(cols_or_exprs) == str:
        cols_or_exprs = [cols_or_exprs]

    # When considering multiple columns,
    # Need an expression which returns null if any part is null else returns concat_ws

    # This is more relevant for our use case e.g. the blocking
    # Since comparisons are only created where blocking values are non null
    null_exprs = [f"{c} is null" for c in cols_or_exprs]
    null_exprs = " OR ".join(null_exprs)
    ws_cols = ", ".join(cols_or_exprs)

    # This is the expression with the correct property
    case_expr = f"""
    case when
    {null_exprs} then null
    else
    concat_ws(', ', {ws_cols})
    end
    """

    # Group name will be the value which will identify this group in the final table
    group_name = _group_name(cols_or_exprs)

    sql = f"""
    (
    select
        count(*) as value_count,

        {case_expr} as value,
        "{group_name}" as group_name,
        (select count({case_expr}) from {table_name}) as total_non_null_rows,
        (select count(*) from {table_name}) as total_rows_inc_nulls,
        (select count(distinct {case_expr}) from {table_name}) as distinct_value_count

    from {table_name}
    where {case_expr} is not null
    group by {case_expr}
    order by group_name, count(*) desc
    )
    """

    return sql


def _array_group(col, df, spark):
    """
    Generate a sql expression that will yield a table with value counts grouped by the provided
    column expression e.g. "dmetaphone(name)" or combination of column expression
    (e.g. ["dmetaphone(first_name)" , "dmetaphone(surname)"]..

    Where an array is provided, value counts will be grouped by the concatenation of
    all the column expressions


    """

    # Group name will be the value which will identify this group in the final table
    group_name = _group_name([col])
    df.createOrReplaceTempView("df")
    sql = f"""
    with df_exp as (
    select
        explode(ifnull({col}, array(null))) as value,
        "{group_name}" as group_name
    from df
    )

    select
        count(*) as value_count,
        value,
        group_name,
        (select count(value) from df_exp) as total_non_null_rows,
        (select count(*) from df_exp) as total_rows_inc_nulls,
        (select count(distinct value) from df_exp) as distinct_value_count
    from df_exp
    where value is not null
    group by value, group_name
    order by group_name, count(*) desc
    """

    return spark.sql(sql)


def _generate_df_all_column_value_frequencies(list_of_col_exprs, df, spark):
    """
    Each element in list_of_col_exprs can be a list or a string
        If string, the contents should be a sql expression which
        is used in the group by to find value frequencies

        If a list, each item in the list is a sql expression which
        are then concatenated using concat_ws, and this is used
        in the gropu by to find value frequencies

    Example:
    [
    'first_name',
    ['first_name', 'surname'],
    ['dmetaphone(first_name), dmetaphone(surname)']
    ]
    """

    df.createOrReplaceTempView("df")

    to_union = [_non_array_group(c, "df") for c in list_of_col_exprs]
    sql = "\n union all \n".join(to_union)
    return spark.sql(sql)


def _generate_df_all_column_value_frequencies_array(list_of_array_cols, df, spark):

    to_union = [_array_group(c, df, spark) for c in list_of_array_cols]
    df = reduce(DataFrame.unionAll, to_union)
    return df


def _get_df_percentiles(df_all_column_value_frequencies, spark):
    """Take df_all_column_value_frequencies and
    turn it into the raw data needed for the percentile cahrt
    """

    df_all_column_value_frequencies.createOrReplaceTempView(
        "df_all_column_value_frequencies"
    )

    sql = """
    select sum(value_count) as sum_tokens_in_value_count_group, value_count, group_name,
    first(total_non_null_rows) as total_non_null_rows,
    first(total_rows_inc_nulls) as total_rows_inc_nulls,
    first(distinct_value_count) as distinct_value_count
    from df_all_column_value_frequencies
    group by group_name, value_count
    order by group_name, value_count desc
    """
    df_total_in_value_counts = spark.sql(sql)
    df_total_in_value_counts.createOrReplaceTempView("df_total_in_value_counts")

    sql = """
    select sum(sum_tokens_in_value_count_group) over (partition by group_name order by value_count desc) as value_count_cumsum,
    sum_tokens_in_value_count_group, value_count, group_name,  total_non_null_rows, total_rows_inc_nulls, distinct_value_count
    from df_total_in_value_counts


    """
    df_total_in_value_counts_cumulative = spark.sql(sql)
    df_total_in_value_counts_cumulative.createOrReplaceTempView(
        "df_total_in_value_counts_cumulative"
    )

    sql = """
    select
    1 - (value_count_cumsum/total_non_null_rows) as percentile_ex_nulls,
    1 - (value_count_cumsum/total_rows_inc_nulls) as percentile_inc_nulls,

    value_count, group_name, total_non_null_rows, total_rows_inc_nulls,
    sum_tokens_in_value_count_group, distinct_value_count
    from df_total_in_value_counts_cumulative

    """
    df_percentiles = spark.sql(sql)
    return df_percentiles


def _get_df_top_bottom_n(
    df_all_column_value_frequencies, spark, limit=20, value_order="desc"
):
    """Take df_all_column_value_frequencies and
    use limit statements to take only the top n values
    """
    df_all_column_value_frequencies.createOrReplaceTempView(
        "df_all_column_value_frequencies"
    )

    sql = """
    select distinct group_name
    from df_all_column_value_frequencies
    """
    group_names = [r.group_name for r in spark.sql(sql).collect()]

    sql = """
    (select *
    from df_all_column_value_frequencies
    where group_name = '{group_name}'
    order by value_count {value_order}
    limit {limit})
    """

    to_union = [
        sql.format(group_name=g, limit=limit, value_order=value_order)
        for g in group_names
    ]

    sql = "\n union all \n".join(to_union)

    df = spark.sql(sql)

    return df


def _collect_and_group_percentiles_df(df_percentiles):
    """Turn df_percentiles into a grouped dictionary
    containing the data need for our charts
    """

    percentiles_groups = {}
    percentiles_rows = [r.asDict() for r in df_percentiles.collect()]
    for r in percentiles_rows:

        if r["group_name"] not in percentiles_groups:

            percentiles_groups[r["group_name"]] = []

            # If we do not already have a row for 100th percentile
            if r["percentile_ex_nulls"] != 1.0:
                first_row = deepcopy(r)
                first_row["percentile_inc_nulls"] = 1.0
                first_row["percentile_ex_nulls"] = 1.0

                percentiles_groups[r["group_name"]].append(first_row)

        percentiles_groups[r["group_name"]].append(r)
        del r["group_name"]
        r["proportion_of_non_null_rows"] = (
            r["sum_tokens_in_value_count_group"] / r["total_non_null_rows"]
        )

    return percentiles_groups


def _collect_and_group_top_values(df_top):
    top_n_groups = {}
    top_n_rows = [r.asDict() for r in df_top.collect()]
    for r in top_n_rows:
        if r["group_name"] not in top_n_groups:
            top_n_groups[r["group_name"]] = []

        top_n_groups[r["group_name"]].append(r)
        del r["group_name"]

    return top_n_groups


def column_value_frequencies_chart(list_of_columns, df, spark, top_n=20, bottom_n=10):
    """Produce value frequency charts for the provided list of column names

    Args:
        list_of_col_combinations (list): A list of column names
        df (DataFrame): Dataframe to profile
        spark (SparkSession): SparkSession object
        top_n (int, optional): Number of values with the highest frequencies to display. Defaults to 20.
        bottom_n (int, optional): Number of values with the lowest frequencies to display. Defaults to 10.

    Returns:
        Chart: If Altair is installed, return a chart. If not,  then it returns the
        vega lite chart spec as a dictionary
    """
    return column_combination_value_frequencies_chart(
        list_of_columns, df, spark, top_n, bottom_n
    )


def column_combination_value_frequencies_chart(
    list_of_col_combinations, df, spark, top_n=20, bottom_n=10
):
    """Produce value frequency charts for the provided list of column names, expressions, or
    combinations thereof.

    Each element in list_of_col_combinations can be a list or a string:
        If string, the contents should be a sql expression which
        is used in the group by to find value frequencies

        If a list, each item in the list is a sql expression which
        are then concatenated using concat_ws, and this is used
        in the gropu by to find value frequencies

    Example:
    list_of_col_combinations = [
        'first_name',
        ['first_name', 'surname'],
        ['dmetaphone(first_name), dmetaphone(surname)']
    ]

    Args:
        list_of_col_combinations (list): A list of column names, expressions or
          combinations thereof for columns of array type
        df (DataFrame): Dataframe to profile
        spark (SparkSession): SparkSession object
        top_n (int, optional): Number of values with the highest frequencies to display. Defaults to 20.
        bottom_n (int, optional): Number of values with the lowest frequencies to display. Defaults to 10.

    Returns:
        Chart: If Altair is installed, return a chart. If not,  then it returns the
        vega lite chart spec as a dictionary
    """
    df_acvf = _generate_df_all_column_value_frequencies(
        list_of_col_combinations, df, spark
    )
    df_acvf.persist()

    df_perc = _get_df_percentiles(df_acvf, spark)
    df_top_n = _get_df_top_bottom_n(df_acvf, spark, top_n)
    df_bottom_n = _get_df_top_bottom_n(df_acvf, spark, bottom_n, value_order="asc")

    df_perc_collected = _collect_and_group_percentiles_df(df_perc)
    df_top_n_collected = _collect_and_group_top_values(df_top_n)
    df_bottom_n_collected = _collect_and_group_top_values(df_bottom_n)

    inner_charts = []
    for col_name in df_top_n_collected.keys():
        top_n_rows = df_top_n_collected[col_name]
        bottom_n_rows = df_bottom_n_collected[col_name]

        percentile_rows = df_perc_collected[col_name]
        inner_chart = _get_inner_chart_spec_freq(
            percentile_rows, top_n_rows, bottom_n_rows, col_name
        )
        inner_charts.append(inner_chart)

    outer_spec = deepcopy(_outer_chart_spec_freq)

    outer_spec["vconcat"] = inner_charts

    if altair_installed:
        return alt.Chart.from_dict(outer_spec)
    else:
        return outer_spec


def array_column_value_frequencies_chart(
    list_of_array_cols: list, df: DataFrame, spark: SparkSession, top_n=20, bottom_n=10
):
    """Produce value frequency charts for the provided list of columns names

    Args:
        list_of_array_cols (list): A list of column names for columns of array type
        df (DataFrame): Dataframe to profile
        spark (SparkSession): SparkSession object
        top_n (int, optional): Number of values with the highest frequencies to display. Defaults to 20.
        bottom_n (int, optional): Number of values with the lowest frequencies to display. Defaults to 10.

    Returns:
        Chart: If Altair is installed, return a chart. If not,  then it returns the
        vega lite chart spec as a dictionary
    """
    df_acvf = _generate_df_all_column_value_frequencies_array(
        list_of_array_cols, df, spark
    )
    df_acvf.persist()

    df_perc = _get_df_percentiles(df_acvf, spark)
    df_top_n = _get_df_top_bottom_n(df_acvf, spark, top_n)
    df_bottom_n = _get_df_top_bottom_n(df_acvf, spark, bottom_n, value_order="asc")

    df_perc_collected = _collect_and_group_percentiles_df(df_perc)
    df_top_n_collected = _collect_and_group_top_values(df_top_n)
    df_bottom_n_collected = _collect_and_group_top_values(df_bottom_n)

    inner_charts = []
    for col_name in df_top_n_collected.keys():
        top_n_rows = df_top_n_collected[col_name]
        bottom_n_rows = df_bottom_n_collected[col_name]
        percentile_rows = df_perc_collected[col_name]
        inner_chart = _get_inner_chart_spec_freq(
            percentile_rows, top_n_rows, bottom_n_rows, col_name
        )
        inner_charts.append(inner_chart)

    outer_spec = deepcopy(_outer_chart_spec_freq)

    outer_spec["vconcat"] = inner_charts

    if altair_installed:
        return alt.Chart.from_dict(outer_spec)
    else:
        return outer_spec


def _parse_blocking_rule(rule):
    rule = rule.lower()
    rule = rule.strip()
    rule = re.sub(r"\s+", " ", rule)
    rule = re.sub(r"\s+", " ", rule)

    cols = rule.split(" and ")

    cols = [[cc.strip() for cc in c.split("=")] for c in cols]

    # check symmetry, return none if not symmetric
    for c in cols:
        left = c[0].replace("r.", "l.")
        right = c[1].replace("r.", "l.")
        if left != right:
            return None

    cols = [c[0].replace("r.", "").replace("l.", "") for c in cols]

    return cols


def blocking_rules_to_column_combinations(blocking_rules: list):
    """Convert blocking rules as specified in a Splink settings dictionary
    into a list of column combintions, the format needed to input into
    column_combination_value_frequencies_chart

    Args:
        blocking_rules (list): A list of blocking rules as specified in a Splink
            settings dictionary

    Returns:
        list: list of column combinations
    """

    column_combinations = [_parse_blocking_rule(r) for r in blocking_rules]
    column_combinations = [c for c in column_combinations if c is not None]
    return column_combinations


def value_frequencies_chart_from_blocking_rules(
    blocking_rules: list, df: DataFrame, spark: SparkSession, top_n=20, bottom_n=10
):
    """Produce value frequency charts for the provided blocking rules

    Args:
        blocking_rules (list): A list of blocking rules as specified in a Splink
            settings dictionary
        df (DataFrame): Dataframe to profile
        spark (SparkSession): SparkSession object
        top_n (int, optional): Number of values with the highest frequencies to display. Defaults to 20.
        bottom_n (int, optional): Number of values with the lowest frequencies to display. Defaults to 10.

    Returns:
        Chart: If Altair is installed, return a chart. If not,  then it returns the
        vega lite chart spec as a dictionary
    """
    col_combinations = blocking_rules_to_column_combinations(blocking_rules)
    return column_combination_value_frequencies_chart(
        col_combinations, df, spark, top_n, bottom_n
    )



# https://stackoverflow.com/questions/52556798/spark-iterative-recursive-algorithms-breaking-spark-lineage
# From https://github.com/high-performance-spark/high-performance-spark-examples/blob/f02142bebf528437702ec8fa689c9c0263e96fe7/high_performance_pyspark/SQLLineage.py#L20
from pyspark.sql.dataframe import DataFrame


def cutLineage(df):
    """
    Cut the lineage of a DataFrame - used for iterative algorithms

    .. Note: This uses internal members and may break between versions
    >>> df = rdd.toDF()
    >>> cutDf = cutLineage(df)
    >>> cutDf.count()
    3
    """
    jRDD = df._jdf.toJavaRDD()
    jSchema = df._jdf.schema()
    jRDD.cache()
    sqlCtx = df.sql_ctx
    try:
        javaSqlCtx = sqlCtx._jsqlContext
    except:
        javaSqlCtx = sqlCtx._ssql_ctx
    newJavaDF = javaSqlCtx.createDataFrame(jRDD, jSchema)
    newDF = DataFrame(newJavaDF, sqlCtx)
    return newDF


def default_break_lineage_blocked_comparisons(df_gammas, spark):
    df_gammas = cutLineage(df_gammas)
    df_gammas.persist()
    return df_gammas



from textwrap import dedent
import sqlglot


def _format_sql(sql):
    try:
        parsed_list = sqlglot.parse(sql, read="spark")
        return_sql = [p.sql(dialect="spark", pretty=True) for p in parsed_list]
        return "\n".join(return_sql)

    except:
        return dedent(sql)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, count, when

from .charts import load_chart_definition, altair_if_installed_else_json

import pandas as pd


def missingness_chart(df: DataFrame):
    """Produce bar chart of missingness in standardised nodes
    Args:
        df (DataFrame): Input Spark dataframe
    Returns:
        Bar chart of missingness
    """
    # Load JSON definition of missingness chart
    chart_path = "missingness_chart_def.json"
    missingness_chart_def = load_chart_definition(chart_path)

    # Data for plot
    # Count and percentage of nulls in each columns as pandas dataframe
    df_nulls = df.select([count(when(col(c).isNull(), c)).alias(c) for c in df.columns])
    pd_nulls = df_nulls.toPandas()
    pd_nulls = pd.melt(pd_nulls)

    record_count = df.count()
    pd_nulls["percentage"] = pd_nulls["value"] / record_count

    # Add data to JSON chart definition
    missingness_chart_def["data"]["values"] = pd_nulls.to_dict("records")

    # Update chart title
    for c in missingness_chart_def["layer"]:
        c["title"] = f"Missingness per column out of {record_count:,.0f} records"

    return altair_if_installed_else_json(missingness_chart_def)



from collections import OrderedDict
import logging


from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from typeguard import typechecked
from .logging_utils import _format_sql
from .settings import ComparisonColumn
from .default_settings import complete_settings_dict
from .ordered_set import OrderedSet
from .parse_case_statement import generate_sql_from_parsed_case_expr

logger = logging.getLogger(__name__)


def _add_left_right(columns_to_retain, name):
    columns_to_retain.add(name + "_l")
    columns_to_retain.add(name + "_r")
    return columns_to_retain


def _add_unique_id_and_source_dataset(
    cols_set: OrderedSet, uid_name: str, sd_name: str, retain_source_dataset_col: bool
):
    if retain_source_dataset_col:
        cols_set.add(f"{sd_name}_l")
        cols_set.add(f"{uid_name}_l")
        cols_set.add(f"{sd_name}_r")
        cols_set.add(f"{uid_name}_r")
    else:
        cols_set.add(f"{uid_name}_l")
        cols_set.add(f"{uid_name}_r")
    return cols_set


def _get_select_expression_gammas(
    settings: dict, retain_source_dataset_col: bool, retain_tf_cols: bool
):
    """Get a select expression which picks which columns to keep in df_gammas

    Args:
        settings (dict): A `splink` settings dictionary
        retain_source_dataset: whether to retain the source dataset columns
        retain_tf_cols: whether to retain the individual term frequency columns

    Returns:
        str: A select expression
    """

    # Use ordered dict as an ordered set - i.e. to make sure we don't have duplicate cols to retain

    select_columns = OrderedSet()

    uid = settings["unique_id_column_name"]
    sds = settings["source_dataset_column_name"]
    select_columns = _add_unique_id_and_source_dataset(
        select_columns, uid, sds, retain_source_dataset_col
    )

    for col in settings["comparison_columns"]:
        cc = ComparisonColumn(col)
        if settings["retain_matching_columns"]:
            for col_name in cc.columns_used:
                select_columns = _add_left_right(select_columns, col_name)
        if col["term_frequency_adjustments"]:
            if retain_tf_cols:
                select_columns = _add_left_right(select_columns, f"tf_{cc.name}")
        case_expr = generate_sql_from_parsed_case_expr(
            col["comparison_levels"], cc.name
        )
        select_columns.add(case_expr)

    for c in settings["additional_columns_to_retain"]:
        select_columns = _add_left_right(select_columns, c)

    if "blocking_rules" in settings:
        if len(settings["blocking_rules"]) > 0:
            select_columns.add("match_key")

    return ", ".join(select_columns)


def _retain_source_dataset_column(settings_dict, df):
    # Want to retain source dataset column in all cases
    # except when link type is dedupe only
    # and the column does not exist in the data
    if settings_dict["link_type"] != "dedupe_only":
        return True

    source_dataset_colname = settings_dict.get(
        "source_dataset_column_name", "source_dataset"
    )
    if f"{source_dataset_colname}_l" in df.columns:
        return True
    else:
        return False


def _retain_tf_columns(settings_dict, df):
    # If all necessary TF columns are in the data,
    # make sure they are retained
    tf_cols = [
        f"tf_{cc['col_name']}" if "col_name" in cc else f"tf_{cc['custom_name']}"
        for cc in settings_dict["comparison_columns"]
        if cc["term_frequency_adjustments"]
    ]

    cols = OrderedSet()
    [_add_left_right(cols, c) for c in tf_cols]

    return all([col in df.columns for col in cols])


def _sql_gen_add_gammas(
    settings: dict,
    df_comparison: DataFrame,
    table_name: str = "df_comparison",
):
    """Build SQL statement that adds gamma columns to the comparison dataframe

    Args:
        settings (dict): `splink` settings dict
        unique_id_col (str, optional): Name of the unique id column. Defaults to "unique_id".
        table_name (str, optional): Name of the comparison df. Defaults to "df_comparison".

    Returns:
        str: A SQL string
    """

    retain_source_dataset = _retain_source_dataset_column(settings, df_comparison)
    retain_tf_cols = _retain_tf_columns(settings, df_comparison)
    select_cols_expr = _get_select_expression_gammas(
        settings, retain_source_dataset, retain_tf_cols
    )

    sql = f"""
    select {select_cols_expr}
    from {table_name}
    """

    return sql


@typechecked
def add_gammas(
    df_comparison: DataFrame,
    settings_dict: dict,
    spark: SparkSession,
    unique_id_col: str = "unique_id",
):
    """Compute the comparison vectors and add them to the dataframe.  See
    https://imai.fas.harvard.edu/research/files/linkage.pdf for more details of what is meant by comparison vectors

    Args:
        df_comparison (spark dataframe): A Spark dataframe containing record comparisons, with records compared using the convention col_name_l, col_name_r
        settings_dict (dict): The `splink` settings dictionary
        spark (Spark session): The Spark session object

    Returns:
        Spark dataframe: A dataframe containing new columns representing the gammas of the model
    """

    settings_dict = complete_settings_dict(settings_dict, spark)

    sql = _sql_gen_add_gammas(settings_dict, df_comparison)

    logger.debug(_format_sql(sql))
    df_comparison.createOrReplaceTempView("df_comparison")
    df_gammas = spark.sql(sql)

    return df_gammas



from pyspark.sql.dataframe import DataFrame
from typing import Union, List
from .model import Model

from pyspark.sql.functions import expr

from splink.vertically_concat import (
    vertically_concatenate_datasets,
)

graphframes_installed = True
try:
    from graphframes import GraphFrame
except ImportError:
    graphframes_installed = False


graphframes_py_missing_message = """
The python package `graphframes` is required to use this function
but is not installed.

For Spark 2.4.5, you can `pip install graphframes=0.6.0`
or download from https://github.com/graphframes/graphframes/tags

For Spark >=3.0.0, the package version you need is not
available from PyPi, and you should download it from
https://github.com/graphframes/graphframes/releases
"""

graphframes_jar_missing_message = """
To use this function, Spark needs access to the `graphframes` jar.

For Spark 2.4.5 suggested code is:

from pyspark.sql import SparkSession

spark = (SparkSession
   .builder
   .appName("my_app")
   .config('spark.jars.packages', 'graphframes:graphframes:0.6.0-spark2.3-s_2.11')
   .getOrCreate()
   )

spark.sparkContext.setCheckpointDir("graphframes_tempdir/")

# Alternatively if no internet access you need to download the jar AND its dependencies and point Spark to the
# location of the jars in your filesystem:
# config('spark.driver.extraClassPath', 'jars/graphframes-0.6.0-spark2.3-s_2.11.jar,jars/scala-logging-api_2.11-2.1.2.jar,jars/scala-logging-slf4j_2.11-2.1.2.jar') # Spark 2.x only
# config('spark.jars', 'jars/graphframes-0.6.0-spark2.3-s_2.11.jar,jars/scala-logging-api_2.11-2.1.2.jar,jars/scala-logging-slf4j_2.11-2.1.2.jar')

Note extraClassPath is needed on spark 2.x only.

You can find these jars, for example, here https://github.com/moj-analytical-services/splink_graph/tree/master/jars

You can find a list of jars corresponding to different versions of Spark here:
https://mvnrepository.com/artifact/graphframes/graphframes?repo=spark-packages

More info on adding jars to Spark here:
https://spark.apache.org/docs/latest/configuration.html#runtime-environment
"""


def _check_graphframes_installation(spark):

    if not graphframes_installed:
        raise Exception(graphframes_py_missing_message)

    all_config = spark.sparkContext.getConf().getAll()

    config_keys = [
        "spark.submit.pyFiles",
        "spark.repl.local.jars",
        "spark.jars.packages",
        "spark.jars",
        "spark.app.initial.file.urls",
        "spark.files",
        "spark.app.initial.jar.urls",
    ]

    graphframe_jar_registered = False
    for (key, value) in all_config:
        if key in config_keys:
            if "graphframes" in value:
                graphframe_jar_registered = True
                
    databricks_backend = False
    for key, value in all_config:
        if "databricks" in key:
            databricks_backend = True

    if not graphframe_jar_registered:
        if not databricks_backend:
            raise Exception(graphframes_jar_missing_message)

    from pyspark.sql import Row

    errored = False
    try:
        n = spark.createDataFrame([Row(id=1)])
        e = spark.createDataFrame([Row(src=1, dst=2)])
        g = GraphFrame(n, e)
        cc = g.connectedComponents()
    except Exception as e:
        error_string = e.__str__()
        errored = True
    if errored:
        raise Exception(
            "There's something wrong with GraphFrames\n"
            "Graphframes is an external library you need to \n"
            "install \n"
            "See quickstart here: https://graphframes.github.io/graphframes/docs/_site/quick-start.html \n"
            "Either it's not installed or it is not \n"
            "working as expected \n\n" + error_string
        )


def _colname_from_threshold(threshold_value):
    v = threshold_value
    return f"cluster_{v:.5g}".replace(".", "_").replace("-", "neg")


def _threshold_values_to_dict(threshold_values):
    if type(threshold_values) is dict:
        return threshold_values
    if type(threshold_values) is list:
        res = {}
        for v in threshold_values:
            res[_colname_from_threshold(v)] = v
        return res

    if type(threshold_values) is float:
        res = {}
        v = threshold_values
        res[_colname_from_threshold(v)] = v
        return res


def clusters_at_thresholds(
    df_of_dfs_nodes: Union[DataFrame, List[DataFrame]],
    df_edges: DataFrame,
    threshold_values: Union[float, list, dict],
    model: Model,
    join_node_details: bool = True,
    check_graphframes_installation: bool = True,
    score_colname: str = "match_probability",
):
    """Generated a table of clusters at one or more threshold_values
    from a table of scored edges (scored pairwise comparisons)

    Args:
        df_of_dfs_nodes (Union[DataFrame, List[DataFrame]]): Dataframe or Dataframes of nodes (original records
            from which pairwise comparisons  are derived).  If the link_type is `dedupe_only`, this will be a
            single dataframe.  If the link_type is `link_and_dedupe` or `link_only`, this will be a list of dataframes.
            The provided dataframes should be the same as provided to Splink().
        df_edges (DataFrame): Dataframe of edges (pairwise record comparisons with scores)
        threshold_values (Union[float, list, dict]): Threshold values of the match probability (or score_colname)
            above which pairwise comparisons are considered to be a match.  There are three options:
            1. A single float value.  Cluster colname will be 'cluster'
            2. An array of float values.  Clustering will be compled for each value, with colname cluster_0_99
            for a threshold value of 0.99 etc.
            3. A dictionary of threshold values.  Dictionary keys will be used for the cluster columns names.
        model (Model): The Splink Model object
        join_node_details (bool, optional): If true, retain the columns from df_nodes in cluster table.
            Defaults to True.
        check_graphframes_installation (bool, optional): Perform checks to see if the graphframes
            installation has been completed before execution. Defaults to True.
        score_colname (str, optional): Score colname. Defaults to "match_probability".  Could also be e.g.
            'match_weight'

    Returns:
        DataFrame: clustered DataFrame
    """

    # dfs is a list of dfs irrespective of whether input was a df or list of dfs
    if type(df_of_dfs_nodes) == DataFrame:
        dfs = [df_of_dfs_nodes]
    else:
        dfs = df_of_dfs_nodes

    spark = dfs[0].sql_ctx.sparkSession
    df_nodes = vertically_concatenate_datasets(dfs)

    if check_graphframes_installation:
        _check_graphframes_installation(spark)

    # Convert threshold_values to a dictionary
    threshold_values = _threshold_values_to_dict(threshold_values)

    # Define a unique id column
    settings_obj = model.current_settings_obj

    uid_colname = settings_obj["unique_id_column_name"]

    if settings_obj["link_type"] == "dedupe_only":
        uid_df_nodes_col = uid_colname
        uid_col_l = f"{uid_colname}_l"
        uid_col_r = f"{uid_colname}_r"
    else:
        source_ds_colname = settings_obj["source_dataset_column_name"]
        uid_df_nodes_col = f"concat({source_ds_colname}, '-__-',{uid_colname})"
        uid_col_l = f"concat({source_ds_colname}_l, '-__-',{uid_colname}_l)"
        uid_col_r = f"concat({source_ds_colname}_r, '-__-',{uid_colname}_r)"

    df_nodes.createOrReplaceTempView("df_nodes")
    df_edges.createOrReplaceTempView("df_edges")

    sql = f"""
    select {uid_df_nodes_col} as id
    from df_nodes
    """
    df_nodes_id = spark.sql(sql)

    cc_thresholds = {}
    for colname, value in threshold_values.items():
        sql = f"""
        select
            {uid_col_l} as src,
            {uid_col_r} as dst
        from df_edges
        where {score_colname} > {value}
        """
        edges_above_thres = spark.sql(sql)
        g = GraphFrame(df_nodes_id, edges_above_thres)
        cc = g.connectedComponents()
        cc_thresholds[colname] = cc

    for cc_col_name, cc in cc_thresholds.items():
        df_nodes_id = df_nodes_id.join(cc, on=["id"], how="left")
        df_nodes_id = df_nodes_id.withColumnRenamed("component", cc_col_name)

    cluster_colnames = cc_thresholds.keys()
    if join_node_details:
        df_nodes_id.createOrReplaceTempView("df_nodes_id")

        df_nodes = df_nodes.withColumn("___id__", expr(uid_df_nodes_col))
        df_nodes.createOrReplaceTempView("df_nodes")

        cluster_sel = ", ".join(cluster_colnames)

        sql = f"""
        select {cluster_sel}, df_nodes.*
        from df_nodes
        left join df_nodes_id
        on df_nodes_id.id = df_nodes.___id__

        """
        df_nodes = spark.sql(sql)
        df_nodes = df_nodes.drop("___id__")
    else:
        df_nodes = df_nodes_id

    return df_nodes



from copy import deepcopy
import os
import json

from pyspark.sql.session import SparkSession

from .settings import Settings
from .default_settings import complete_settings_dict
from .charts import (
    load_chart_definition,
    altair_if_installed_else_json,
    _load_multi_chart_template,
    _load_external_libs,
    _make_json,
)

import logging

logger = logging.getLogger(__name__)


class Model:
    """Stores the current model parameters (in self.params) and values for params for all previous iterations

    Attributes:
        params (str): A dictionary storing the current parameters.

    """

    def __init__(self, settings: dict, spark: SparkSession):
        """

        Args:
            settings (dict): A splink setting dictionary
            spark (SparkSession): Your sparksession. Defaults to None.
        """

        self.historical_settings = []

        self.iteration = 0

        # Settings is just a wrapper around the settings dict
        settings_dict_original = deepcopy(settings)
        self.original_settings_obj = Settings(settings_dict_original)

        settings_dict_completed = complete_settings_dict(deepcopy(settings), spark)

        self.current_settings_obj = Settings(deepcopy(settings_dict_completed))

    def _populate_model_from_maximisation_step(self, lambda_value, pi_df_collected):
        """
        Take results of sql query that computes updated values
        and update parameters.

        df_pi_collected is like
        gamma_value, new_probability_match, new_probability_non_match, gamma_col
        """

        self.current_settings_obj.reset_all_probabilities()

        self.current_settings_obj["proportion_of_matches"] = lambda_value
        for row_dict in pi_df_collected:
            name = row_dict["column_name"]
            level_int = row_dict["gamma_value"]
            match_prob = row_dict["new_probability_match"]
            ### NOT SURE ABOUT HOW TO DEAL WITH THIS SO FOR NOW I'M JUST MAKING
            ### SURE U-PROBS EXIST FOR ALL LEVELS WITH SOME MINIMAL VALUE...
            non_match_prob = row_dict["new_probability_non_match"] or 1e-10

            self.current_settings_obj.set_m_probability(
                name, level_int, match_prob, force=False
            )
            self.current_settings_obj.set_u_probability(
                name, level_int, non_match_prob, force=False
            )

    def is_converged(self):
        p_latest = self.current_settings_obj
        p_previous = self.historical_settings[-1]
        threshold = self.current_settings_obj["em_convergence"]

        diffs = []

        change_lambda = abs(
            p_latest["proportion_of_matches"] - p_previous["proportion_of_matches"]
        )
        diffs.append(
            {"col_name": "proportion_of_matches", "diff": change_lambda, "level": ""}
        )

        compare = zip(
            p_latest.comparison_columns_list, p_previous.comparison_columns_list
        )
        for c_latest, c_previous in compare:
            for m_or_u in ["m_probabilities", "u_probabilities"]:
                for gamma_index in range(c_latest.num_levels):
                    val_latest = c_latest[m_or_u][gamma_index]
                    val_previous = c_previous[m_or_u][gamma_index]
                    if val_latest is not None:
                        diff = abs(val_latest - val_previous)
                    else:
                        diff = 0
                    diffs.append(
                        {"col_name": c_latest.name, "diff": diff, "level": gamma_index}
                    )

        diffs = sorted(diffs, key=lambda x: x["diff"], reverse=True)
        largest_diff = diffs[0]["diff"]
        largest_diff_name = diffs[0]["col_name"]
        largest_diff_level = diffs[0]["level"]

        if largest_diff_level != "":
            level_info = f", level {largest_diff_level}"
        else:
            level_info = ""
        logger.info(
            f"The maximum change in parameters was {largest_diff} for key {largest_diff_name}{level_info}"
        )

        return largest_diff < threshold

    def save_settings_to_iteration_history(self):
        """
        Take current params and
        """
        current_params = deepcopy(self.current_settings_obj.settings_dict)

        self.historical_settings.append(Settings(current_params))

    def _to_dict(self):
        p_dict = {}
        p_dict["current_settings_dict"] = self.current_settings_obj.settings_dict
        p_dict["historical_settings_dicts"] = [
            s.settings_dict for s in self.historical_settings
        ]
        p_dict["original_settings_dict"] = self.original_settings_obj.settings_dict
        p_dict["iteration"] = self.iteration

        return p_dict

    def save_model_to_json_file(self, path, overwrite=False):

        if os.path.isfile(path) and not overwrite:
            raise ValueError(
                f"The path {path} already exists. Please provide a different path."
            )

        d = self._to_dict()
        with open(path, "w") as f:
            json.dump(d, f, indent=4)

    #######################################################################################
    # The rest of this module is just 'presentational' elements - charts, and __repr__ etc.
    #######################################################################################

    def m_u_history_as_rows(self):
        rows = []
        for it_num, p in enumerate(self.historical_settings):
            new_rows = p.m_u_as_rows()
            for r in new_rows:
                r["iteration"] = it_num
                r["final"] = it_num == self.iteration - 1

            rows.extend(new_rows)
        return rows

    def lambda_history_as_rows(self):
        rows = []
        for it_num, p in enumerate(self.historical_settings):
            rows.append({"λ": p["proportion_of_matches"], "iteration": it_num})

        return rows

    def lambda_iteration_chart(self):  # pragma: no cover
        chart_path = "lambda_iteration_chart_def.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.lambda_history_as_rows()
        return altair_if_installed_else_json(chart)

    def gamma_distribution_chart(self):  # pragma: no cover
        chart_path = "gamma_distribution_chart_def.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.current_settings_obj.m_u_as_rows()
        return altair_if_installed_else_json(chart)

    def bayes_factor_chart(self):
        return self.current_settings_obj.bayes_factor_chart()

    def probability_distribution_chart(self):
        return self.current_settings_obj.probability_distribution_chart()

    def bayes_factor_history_charts(self):
        chart_path = "bayes_factor_history_chart_def.json"
        chart_template = load_chart_definition(chart_path)

        # Empty list of chart definitions
        chart_defs = []

        # Full iteration history
        data = self.m_u_history_as_rows()

        # Create charts for each column
        for cc in self.current_settings_obj.comparison_columns_list:

            chart_def = deepcopy(chart_template)

            # Assign iteration history to values of chart_def
            chart_def["data"]["values"] = [
                d for d in data if d["column_name"] == cc.name
            ]
            chart_def["title"]["text"] = cc.name
            chart_def["hconcat"][1]["layer"][0]["encoding"]["color"]["legend"][
                "tickCount"
            ] = (cc.num_levels - 1)
            chart_defs.append(chart_def)

        combined_charts = {
            "config": {
                "view": {"width": 400, "height": 120},
            },
            "title": {"text": "Bayes factor iteration history", "anchor": "middle"},
            "vconcat": chart_defs,
            "resolve": {"scale": {"color": "independent"}},
            "$schema": "https://vega.github.io/schema/vega-lite/v4.json",
        }

        return altair_if_installed_else_json(combined_charts)

    def m_u_probabilities_iteration_chart(self):  # pragma: no cover
        chart_path = "m_u_iteration_chart_def.json"
        chart = load_chart_definition(chart_path)
        chart["data"]["values"] = self.m_u_history_as_rows()
        return altair_if_installed_else_json(chart)

    def all_charts_write_html_file(
        self, filename="splink_charts.html", overwrite=False
    ):

        if os.path.isfile(filename) and not overwrite:
            raise ValueError(
                f"The path {filename} already exists. Please provide a different path."
            )

        template = _load_multi_chart_template()
        fmt_dict = _load_external_libs()

        fmt_dict["c_prob_dist"] = _make_json(self.probability_distribution_chart())
        fmt_dict["c_bayes_factor"] = _make_json(self.bayes_factor_chart())
        fmt_dict["c_gamma_dist"] = _make_json(self.gamma_distribution_chart())
        fmt_dict["c_lambda_it"] = _make_json(self.lambda_iteration_chart())
        fmt_dict["c_bayes_factor_hist"] = _make_json(self.bayes_factor_history_charts())
        fmt_dict["c_m_u_hist"] = _make_json(self.m_u_probabilities_iteration_chart())
        fmt_dict["proportion_of_matches"] = self.current_settings_obj[
            "proportion_of_matches"
        ]
        with open(filename, "w") as f:
            f.write(template.format(**fmt_dict))

    def _repr_pretty_(self, p, cycle):  # pragma: no cover

        return p.pretty(self.current_settings_obj)


def load_model_from_json(path):
    # Load params
    with open(path, "r") as f:
        params_from_json = json.load(f)

    p = load_model_from_dict(params_from_json)

    return p


def load_model_from_dict(model_dict):

    keys = set(model_dict.keys())

    expected_keys = {
        "current_settings_dict",
        "historical_settings_dicts",
        "original_settings_dict",
        "iteration",
    }

    if keys == expected_keys:
        p = Model(model_dict["current_settings_dict"], spark=None)

        p.historical_settings = [
            Settings(s) for s in model_dict["historical_settings_dicts"]
        ]
        p.original_settings_obj = Settings(model_dict["original_settings_dict"])
        p.iteration = model_dict["iteration"]
    else:
        raise ValueError("Your saved params seem to be corrupted")

    return p



import re


def get_largest_blocks(blocking_rule, df, spark, limit=5):
    """
    For a given blocking rule, find out which will be the largest blocks
    """

    parts = re.split(" |=", blocking_rule)
    parts = [p for p in parts if "l." in p]
    parts = [p.replace("l.", "") for p in parts]

    col_expr = ", ".join(parts)

    filter_nulls_expr = " and ".join(f"{p} is not null" for p in parts)

    sql = f"""
    SELECT {col_expr}, count(*) as count
    FROM df
    WHERE {filter_nulls_expr}
    GROUP BY {col_expr}
    ORDER BY count(*) desc
    LIMIT {limit}
    """
    df.createOrReplaceTempView("df")
    return spark.sql(sql)



from .model import Model

from .charts import load_chart_definition, altair_if_installed_else_json

import pandas as pd
from math import log2

initial_template = """
Initial probability of match (prior) = λ = {lam:.4g}
"""

col_template = [
    ("Comparison of {column_name}.  Values are:", ""),
    ("{column_name}_l:", "{value_l}"),
    ("{column_name}_r:", "{value_r}"),
    ("Comparison has:", "{num_levels} levels"),
    ("Level for this comparison:", "{gamma_column_name} = {gamma_index}"),
    ("m probability = P(level|match):", "{m_probability:.4g}"),
    ("u probability = P(level|non-match):", "{u_probability:.4g}"),
    ("Bayes factor = m/u:", "{bayes_factor:.4g}"),
    ("New probability of match (updated belief):", "{updated_belief:.4g}"),
]

end_template = """
Final probability of match = {final:.4g}

Reminder:

The m probability for a given level is the proportion of matches which are in this level.
We would generally expect the highest similarity level to have the largest proportion of matches.
For example, we would expect first name field to match exactly amongst most matching records, except where nicknames, aliases or typos have occurred.
For a comparison column that changes through time, like address, we may expect a lower proportion of comparisons to be in the highest similarity level.

The u probability for a given level is the proportion of non-matches which are in this level.
We would generally expect the lowest similarity level to have the highest proportion of non-matches, but the magnitude depends on the cardinality of the field.
For example, we would expect that in the vast majority of non-matching records, the date of birth field would not match.  However, we would expect it to be common for gender to match amongst non-matches.
"""


def intuition_report(row_dict: dict, model: Model):
    """Generate a text summary of a row in the comparison table which explains how the match_probability was computed

    Args:
        row_dict (dict): A python dictionary representing the comparison row
        model (Model): splink Model object

    Returns:
        string: The intuition report
    """

    lam = model.current_settings_obj["proportion_of_matches"]
    report = initial_template.format(lam=lam)
    current_prob = lam

    for cc in model.current_settings_obj.comparison_columns_list:
        d = cc.df_e_row_intuition_dict(row_dict)

        bf = d["bayes_factor"]

        a = bf * current_prob
        new_p = a / (a + (1 - current_prob))
        d["updated_belief"] = new_p
        current_prob = new_p

        col_report = []
        col_report.append("------")
        for (blurb, value) in col_template:
            blurb_fmt = blurb.format(**d)

            value_fmt = value.format(**d)
            col_report.append(f"{blurb_fmt:<50} {value_fmt}")
        col_report.append("\n")
        col_report = "\n".join(col_report)
        report += col_report

    report += end_template.format(final=new_p)

    if len(model.current_settings_obj["blocking_rules"]) > 1:
        match_key = int(row_dict["match_key"])
        br = model.current_settings_obj["blocking_rules"][match_key]
        br = f"\nThis comparison was generated by the blocking rule: {br}"
        report += br

    return report


def _get_bayes_factors(row_dict, model):
    bayes_factors = []

    for cc in model.current_settings_obj.comparison_columns_list:
        row_desc = cc.df_e_row_intuition_dict(row_dict)
        bayes_factors.append(row_desc)

    return bayes_factors


def bayes_factor_intuition_chart(row_dict, model):
    chart_path = "bayes_factor_intuition_chart_def.json"
    bayes_factor_intuition_chart_def = load_chart_definition(chart_path)

    data = _get_bayes_factors(row_dict, model)

    # Get initial and final bayes factors
    lam = model.current_settings_obj["proportion_of_matches"]
    bf_init = lam / (1 - lam)
    bf_final = sum([d["log2_bayes_factor"] for d in data]) + log2(bf_init)

    # Sort records in descending order of influence
    # with start and end positions added
    df = pd.DataFrame(data)
    df["abs_log2_bayes_factor"] = abs(df["log2_bayes_factor"])
    df = (
        df.sort_values(by="abs_log2_bayes_factor", ascending=False)
        .drop("abs_log2_bayes_factor", axis=1)
        .reset_index(drop=True)
        .append(
            {
                "bayes_factor": 2 ** bf_final,
                "log2_bayes_factor": bf_final,
                "column_name": "Final score",
            },
            ignore_index=True,
        )
    )
    df = (
        pd.DataFrame(
            {
                "bayes_factor": bf_init,
                "log2_bayes_factor": log2(bf_init),
                "column_name": "Prior lambda",
            },
            index=[0],
        )
        .append(df, ignore_index=True)
        .reset_index()
    )

    bayes_factor_intuition_chart_def["data"]["values"] = df.to_dict("records")

    return altair_if_installed_else_json(bayes_factor_intuition_chart_def)
